# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from os.path import join

import pyomo.common.unittest as unittest

from pyomo.common.fileutils import import_file, PYOMO_ROOT_DIR
from pyomo.contrib.satsolver.satsolver import satisfiable, z3_available
from pyomo.core.base.set_types import PositiveIntegers, NonNegativeReals, Binary
from pyomo.environ import (
    ConcreteModel,
    Var,
    Constraint,
    Objective,
    sin,
    cos,
    tan,
    asin,
    acos,
    atan,
    sqrt,
    log,
    minimize,
)
from pyomo.gdp import Disjunct, Disjunction

exdir = join(PYOMO_ROOT_DIR, 'examples', 'gdp')


@unittest.skipUnless(z3_available, "Z3 SAT solver is not available.")
class SatSolverTests(unittest.TestCase):
    def test_simple_sat_model(self):
        m = ConcreteModel()
        m.x = Var()
        m.c = Constraint(expr=1 == m.x)
        m.o = Objective(expr=m.x)
        self.assertTrue(satisfiable(m))

    def test_simple_unsat_model(self):
        m = ConcreteModel()
        m.x = Var()
        m.c1 = Constraint(expr=1 == m.x)
        m.c2 = Constraint(expr=2 == m.x)
        m.o = Objective(expr=m.x)
        self.assertFalse(satisfiable(m))

    def test_bounds_sat(self):
        m = ConcreteModel()
        m.x = Var(bounds=(0, 5))
        m.c1 = Constraint(expr=4.99 == m.x)
        m.o = Objective(expr=m.x)
        self.assertTrue(satisfiable(m))

    def test_upper_bound_unsat(self):
        m = ConcreteModel()
        m.x = Var(bounds=(0, 5))
        m.c = Constraint(expr=5.01 == m.x)
        m.o = Objective(expr=m.x)
        self.assertFalse(satisfiable(m))

    def test_lower_bound_unsat(self):
        m = ConcreteModel()
        m.x = Var(bounds=(0, 5))
        m.c = Constraint(expr=-0.01 == m.x)
        m.o = Objective(expr=m.x)
        self.assertFalse(satisfiable(m))

    def test_binary_expressions(self):
        m = ConcreteModel()
        m.x = Var()
        m.y = Var()
        m.z = Var()
        m.c = Constraint(expr=0 <= m.x + m.y - m.z * m.y / m.x + 7)
        m.o = Objective(expr=m.x)
        self.assertTrue(satisfiable(m))

    def test_unary_expressions(self):
        m = ConcreteModel()
        m.x = Var()
        m.y = Var()
        m.z = Var()
        m.a = Var()
        m.b = Var()
        m.c = Var()
        m.d = Var()
        m.c1 = Constraint(expr=0 <= sin(m.x))
        m.c2 = Constraint(expr=0 <= cos(m.y))
        m.c3 = Constraint(expr=0 <= tan(m.z))
        m.c4 = Constraint(expr=0 <= asin(m.a))
        m.c5 = Constraint(expr=0 <= acos(m.b))
        m.c6 = Constraint(expr=0 <= atan(m.c))
        m.c7 = Constraint(expr=0 <= sqrt(m.d))
        m.o = Objective(expr=m.x)
        self.assertTrue(satisfiable(m) is not False)

    def test_unhandled_expressions(self):
        m = ConcreteModel()
        m.x = Var()
        m.c1 = Constraint(expr=0 <= log(m.x))
        self.assertTrue(satisfiable(m))

    def test_abs_expressions(self):
        m = ConcreteModel()
        m.x = Var()
        m.c1 = Constraint(expr=-0.001 >= abs(m.x))
        m.o = Objective(expr=m.x)
        self.assertFalse(satisfiable(m))

    def test_inactive_constraints(self):
        m = ConcreteModel()
        m.x = Var()
        m.c1 = Constraint(expr=m.x == 1)
        m.c2 = Constraint(expr=m.x == 2)
        m.o = Objective(expr=m.x)
        self.assertFalse(satisfiable(m))
        m.c2.deactivate()
        self.assertTrue(satisfiable(m))

    def test_disjunction_sat1(self):
        m = ConcreteModel()
        m.x1 = Var(bounds=(0, 8))
        m.x2 = Var(bounds=(0, 8))
        m.obj = Objective(expr=m.x1 + m.x2, sense=minimize)
        m.y1 = Disjunct()
        m.y2 = Disjunct()
        m.y1.c1 = Constraint(expr=m.x1 >= 2)
        m.y1.c2 = Constraint(expr=m.x2 >= 2)
        m.y2.c1 = Constraint(expr=m.x1 >= 9)
        m.y2.c2 = Constraint(expr=m.x2 >= 3)
        m.djn = Disjunction(expr=[m.y1, m.y2])
        self.assertTrue(satisfiable(m))

    def test_disjunction_sat1(self):
        m = ConcreteModel()
        m.x1 = Var(bounds=(0, 8))
        m.x2 = Var(bounds=(0, 8))
        m.obj = Objective(expr=m.x1 + m.x2, sense=minimize)
        m.y1 = Disjunct()
        m.y2 = Disjunct()
        m.y1.c1 = Constraint(expr=m.x1 >= 9)
        m.y1.c2 = Constraint(expr=m.x2 >= 2)
        m.y2.c1 = Constraint(expr=m.x1 >= 3)
        m.y2.c2 = Constraint(expr=m.x2 >= 3)
        m.djn = Disjunction(expr=[m.y1, m.y2])
        self.assertTrue(satisfiable(m))

    def test_disjunction_unsat(self):
        m = ConcreteModel()
        m.x1 = Var(bounds=(0, 8))
        m.x2 = Var(bounds=(0, 8))
        m.obj = Objective(expr=m.x1 + m.x2, sense=minimize)
        m.y1 = Disjunct()
        m.y2 = Disjunct()
        m.y1.c1 = Constraint(expr=m.x1 >= 9)
        m.y1.c2 = Constraint(expr=m.x2 >= 2)
        m.y2.c1 = Constraint(expr=m.x1 >= 3)
        m.y2.c2 = Constraint(expr=m.x2 >= 9)
        m.djn = Disjunction(expr=[m.y1, m.y2])
        self.assertFalse(satisfiable(m))

    def test_multiple_disjunctions_unsat(self):
        m = ConcreteModel()
        m.x1 = Var(bounds=(0, 8))
        m.x2 = Var(bounds=(0, 8))
        m.obj = Objective(expr=m.x1 + m.x2, sense=minimize)
        m.y1 = Disjunct()
        m.y2 = Disjunct()
        m.y1.c1 = Constraint(expr=m.x1 >= 2)
        m.y1.c2 = Constraint(expr=m.x2 >= 2)
        m.y2.c1 = Constraint(expr=m.x1 >= 2)
        m.y2.c2 = Constraint(expr=m.x2 >= 2)
        m.djn1 = Disjunction(expr=[m.y1, m.y2])
        m.z1 = Disjunct()
        m.z2 = Disjunct()
        m.z1.c1 = Constraint(expr=m.x1 <= 1)
        m.z1.c2 = Constraint(expr=m.x2 <= 1)
        m.z2.c1 = Constraint(expr=m.x1 <= 1)
        m.z2.c2 = Constraint(expr=m.x2 <= 1)
        m.djn2 = Disjunction(expr=[m.z1, m.z2])
        self.assertFalse(satisfiable(m))

    def test_multiple_disjunctions_sat(self):
        m = ConcreteModel()
        m.x1 = Var(bounds=(0, 8))
        m.x2 = Var(bounds=(0, 8))
        m.obj = Objective(expr=m.x1 + m.x2, sense=minimize)
        m.y1 = Disjunct()
        m.y2 = Disjunct()
        m.y1.c1 = Constraint(expr=m.x1 >= 2)
        m.y1.c2 = Constraint(expr=m.x2 >= 2)
        m.y2.c1 = Constraint(expr=m.x1 >= 1)
        m.y2.c2 = Constraint(expr=m.x2 >= 1)
        m.djn1 = Disjunction(expr=[m.y1, m.y2])
        m.z1 = Disjunct()
        m.z2 = Disjunct()
        m.z1.c1 = Constraint(expr=m.x1 <= 1)
        m.z1.c2 = Constraint(expr=m.x2 <= 1)
        m.z2.c1 = Constraint(expr=m.x1 <= 0)
        m.z2.c2 = Constraint(expr=m.x2 <= 0)
        m.djn2 = Disjunction(expr=[m.z1, m.z2])
        self.assertTrue(satisfiable(m))

    def test_integer_domains(self):
        m = ConcreteModel()
        m.x1 = Var(domain=PositiveIntegers)
        m.c1 = Constraint(expr=m.x1 == 0.5)
        self.assertFalse(satisfiable(m))

    def test_real_domains(self):
        m = ConcreteModel()
        m.x1 = Var(domain=NonNegativeReals)
        m.c1 = Constraint(expr=m.x1 == -1.3)
        self.assertFalse(satisfiable(m))

    def test_binary_domains(self):
        m = ConcreteModel()
        m.x1 = Var(domain=Binary)
        m.c1 = Constraint(expr=m.x1 == 2)
        self.assertFalse(satisfiable(m))

    def test_8PP(self):
        exfile = import_file(join(exdir, 'eight_process', 'eight_proc_model.py'))
        m = exfile.build_eight_process_flowsheet()
        self.assertTrue(satisfiable(m) is not False)

    def test_8PP_deactivate(self):
        exfile = import_file(join(exdir, 'eight_process', 'eight_proc_model.py'))
        m = exfile.build_eight_process_flowsheet()
        for djn in m.component_data_objects(ctype=Disjunction):
            djn.deactivate()
        self.assertTrue(satisfiable(m) is not False)

    def test_strip_pack(self):
        exfile = import_file(join(exdir, 'strip_packing', 'strip_packing_concrete.py'))
        m = exfile.build_rect_strip_packing_model()
        self.assertTrue(satisfiable(m))

    def test_constrained_layout(self):
        exfile = import_file(join(exdir, 'constrained_layout', 'cons_layout_model.py'))
        m = exfile.build_constrained_layout_model()
        self.assertTrue(satisfiable(m) is not False)

    def test_ex_633_trespalacios(self):
        exfile = import_file(join(exdir, 'small_lit', 'ex_633_trespalacios.py'))
        m = exfile.build_simple_nonconvex_gdp()
        self.assertTrue(satisfiable(m) is not False)


if __name__ == '__main__':
    unittest.main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import logging
from weakref import ref as weakref_ref, ReferenceType

from pyomo.common.deprecation import deprecation_warning, RenamedClass
from pyomo.common.log import is_debug_set
from pyomo.common.modeling import unique_component_name, NOTSET
from pyomo.common.timing import ConstructionTimer
from pyomo.core.staleflag import StaleFlagManager
from pyomo.core.expr.boolean_value import BooleanValue
from pyomo.core.expr import GetItemExpression
from pyomo.core.expr.expr_common import _type_check_exception_arg
from pyomo.core.expr.numvalue import value
from pyomo.core.base.component import ComponentData, ModelComponentFactory
from pyomo.core.base.global_set import UnindexedComponent_index
from pyomo.core.base.indexed_component import IndexedComponent, UnindexedComponent_set
from pyomo.core.base.misc import apply_indexed_rule
from pyomo.core.base.set import Set, BooleanSet, Binary
from pyomo.core.base.util import is_functor
from pyomo.core.base.var import Var

logger = logging.getLogger('pyomo.core')

_logical_var_types = {bool, type(None)}


class _DeprecatedImplicitAssociatedBinaryVariable:
    __slots__ = ('_boolvar',)

    def __init__(self, boolvar):
        self._boolvar = weakref_ref(boolvar)

    def __call__(self):
        deprecation_warning(
            "Relying on core.logical_to_linear to transform "
            "BooleanVars that do not appear in LogicalConstraints "
            "is deprecated. Please associate your own binaries if "
            "you have BooleanVars not used in logical expressions.",
            version='6.2',
        )

        parent_block = self._boolvar().parent_block()
        new_var = Var(domain=Binary)
        parent_block.add_component(
            unique_component_name(
                parent_block, self._boolvar().local_name + "_asbinary"
            ),
            new_var,
        )
        self._boolvar()._associated_binary = None
        self._boolvar().associate_binary_var(new_var)
        return new_var

    def __getstate__(self):
        return self._boolvar()

    def __setstate__(self, state):
        self._boolvar = weakref_ref(state)


def _associated_binary_mapper(encode, val):
    if val is None:
        return None
    if encode:
        if val.__class__ is not _DeprecatedImplicitAssociatedBinaryVariable:
            return val()
    else:
        if val.__class__ is not _DeprecatedImplicitAssociatedBinaryVariable:
            return weakref_ref(val)
    return val


class BooleanVarData(ComponentData, BooleanValue):
    """This class defines the data for a single Boolean variable.

    Parameters
    ----------
    component: Component
        The BooleanVar object that owns this data.

    Attributes
    ----------
    fixed: bool
        If True, then this variable is treated as a fixed constant in
        the model.

    """

    __slots__ = ('_value', 'fixed', '_stale', '_associated_binary')
    __autoslot_mappers__ = {
        '_associated_binary': _associated_binary_mapper,
        '_stale': StaleFlagManager.stale_mapper,
    }

    def __init__(self, component=None):
        #
        # These lines represent in-lining of the
        # following constructors:
        #   - BooleanVarData
        #   - ComponentData
        #   - BooleanValue
        self._component = weakref_ref(component) if (component is not None) else None
        self._index = NOTSET
        self._value = None
        self.fixed = False
        self._stale = 0  # True

        self._associated_binary = None

    def is_fixed(self):
        """Returns True if this variable is fixed, otherwise returns False."""
        return self.fixed

    def is_constant(self):
        """Returns False because this is not a constant in an expression."""
        return False

    def is_variable_type(self):
        """Returns True because this is a variable."""
        return True

    def is_potentially_variable(self):
        """Returns True because this is a variable."""
        return True

    def set_value(self, val, skip_validation=False):
        """
        Set the value of this numeric object, after
        validating its value. If the 'valid' flag is True,
        then the validation step is skipped.
        """
        # Note that it is basically as fast to check the type as it is
        # to check the skip_validation flag.  Considering that we expect
        # the flag to always be False, we will just ignore it in the
        # name of efficiency.
        if val.__class__ not in _logical_var_types:
            if not skip_validation:
                logger.warning(
                    "implicitly casting '%s' value %s to bool" % (self.name, val)
                )
            val = bool(val)
        self._value = val
        self._stale = StaleFlagManager.get_flag(self._stale)

    def clear(self):
        self.value = None

    def __call__(self, exception=NOTSET):
        """Compute the value of this variable."""
        exception = _type_check_exception_arg(self, exception)
        return self.value

    @property
    def value(self):
        """bool : the current value for this variable."""
        return self._value

    @value.setter
    def value(self, val):
        self.set_value(val)

    @property
    def domain(self):
        """BooleanSet : the domain for this variable."""
        return BooleanSet

    @property
    def stale(self):
        """
        bool : A Boolean indicating whether the value of this variable is
        Consistent with the most recent solve.  `True` indicates that
        this variable's value was set prior to the most recent solve and
        was not updated by the results returned by the solve.
        """
        return StaleFlagManager.is_stale(self._stale)

    @stale.setter
    def stale(self, val):
        if val:
            self._stale = 0
        else:
            self._stale = StaleFlagManager.get_flag(0)

    def get_associated_binary(self):
        """Get the binary VarData associated with this
        BooleanVarData"""
        return (
            self._associated_binary() if self._associated_binary is not None else None
        )

    def associate_binary_var(self, binary_var):
        """Associate a binary VarData to this BooleanVarData"""
        if (
            self._associated_binary is not None
            and type(self._associated_binary)
            is not _DeprecatedImplicitAssociatedBinaryVariable
        ):
            raise RuntimeError(
                "Reassociating BooleanVar '%s' (currently associated "
                "with '%s') with '%s' is not allowed"
                % (
                    self.name,
                    (
                        self._associated_binary().name
                        if self._associated_binary is not None
                        else None
                    ),
                    binary_var.name if binary_var is not None else None,
                )
            )
        if binary_var is not None:
            self._associated_binary = weakref_ref(binary_var)

    def fix(self, value=NOTSET, skip_validation=False):
        """Fix the value of this variable (treat as nonvariable)

        This sets the `fixed` indicator to True.  If ``value`` is
        provided, the value (and the ``skip_validation`` flag) are first
        passed to :py:meth:`set_value()`.

        """
        self.fixed = True
        if value is not NOTSET:
            self.set_value(value, skip_validation)

    def unfix(self):
        """Unfix this variable (treat as variable)

        This sets the `fixed` indicator to False.

        """
        self.fixed = False

    def free(self):
        """Alias for :py:meth:`unfix`"""
        return self.unfix()


class _BooleanVarData(metaclass=RenamedClass):
    __renamed__new_class__ = BooleanVarData
    __renamed__version__ = '6.7.2'


class _GeneralBooleanVarData(metaclass=RenamedClass):
    __renamed__new_class__ = BooleanVarData
    __renamed__version__ = '6.7.2'


@ModelComponentFactory.register("Logical decision variables.")
class BooleanVar(IndexedComponent):
    """A logical variable, which may be defined over an index.

    Args:
        initialize (float or function, optional): The initial value for
            the variable, or a rule that returns initial values.
        rule (float or function, optional): An alias for `initialize`.
        dense (bool, optional): Instantiate all elements from
            `index_set()` when constructing the Var (True) or just the
            variables returned by `initialize`/`rule` (False).  Defaults
            to True.
    """

    _ComponentDataClass = BooleanVarData

    def __new__(cls, *args, **kwds):
        if cls != BooleanVar:
            return super(BooleanVar, cls).__new__(cls)
        if not args or (args[0] is UnindexedComponent_set and len(args) == 1):
            return ScalarBooleanVar.__new__(ScalarBooleanVar)
        else:
            return IndexedBooleanVar.__new__(IndexedBooleanVar)

    def __init__(self, *args, **kwd):
        initialize = kwd.pop('initialize', None)
        initialize = kwd.pop('rule', initialize)
        self._dense = kwd.pop('dense', True)

        kwd.setdefault('ctype', BooleanVar)
        IndexedComponent.__init__(self, *args, **kwd)

        # Allow for functions or functors for value initialization,
        # without confusing with Params, etc (which have a __call__ method).
        #
        self._value_init_value = None
        self._value_init_rule = None

        if is_functor(initialize) and (not isinstance(initialize, BooleanValue)):
            self._value_init_rule = initialize
        else:
            self._value_init_value = initialize

    def flag_as_stale(self):
        """
        Set the 'stale' attribute of every variable data object to True.
        """
        for boolvar_data in self._data.values():
            boolvar_data._stale = 0

    def get_values(self, include_fixed_values=True):
        """
        Return a dictionary of index-value pairs.
        """
        if include_fixed_values:
            return dict((idx, vardata.value) for idx, vardata in self._data.items())
        return dict(
            (idx, vardata.value)
            for idx, vardata in self._data.items()
            if not vardata.fixed
        )

    extract_values = get_values

    def set_values(self, new_values, skip_validation=False):
        """
        Set data values from a dictionary.

        The default behavior is to validate the values in the
        dictionary.
        """
        for index, new_value in new_values.items():
            self[index].set_value(new_value, skip_validation)

    def construct(self, data=None):
        """Construct this component."""
        if is_debug_set(logger):  # pragma:nocover
            try:
                name = str(self.name)
            except:
                name = type(self)
            logger.debug(
                "Constructing Variable, name=%s, from data=%s" % (name, str(data))
            )

        if self._constructed:
            return
        timer = ConstructionTimer(self)
        self._constructed = True

        if self._anonymous_sets is not None:
            for _set in self._anonymous_sets:
                _set.construct()

        #
        # Construct BooleanVarData objects for all index values
        #
        if not self.is_indexed():
            self._data[None] = self
            self._initialize_members((None,))
        elif self._dense:
            # This loop is optimized for speed with pypy.
            # Calling dict.update((...) for ...) is roughly
            # 30% slower
            self_weakref = weakref_ref(self)
            for ndx in self._index_set:
                cdata = self._ComponentDataClass(component=None)
                cdata._component = self_weakref
                self._data[ndx] = cdata
                # NOTE: This is a special case where a key, value pair is added
                # to the _data dictionary without calling
                # _getitem_when_not_present, which is why we need to set the
                # index here.
                cdata._index = ndx
            self._initialize_members(self._index_set)
        timer.report()

    def add(self, index):
        """Add a variable with a particular index."""
        return self[index]

    #
    # This method must be defined on subclasses of
    # IndexedComponent that support implicit definition
    #
    def _getitem_when_not_present(self, index):
        """Returns the default component data value."""
        if index is None and not self.is_indexed():
            obj = self._data[index] = self
        else:
            obj = self._data[index] = self._ComponentDataClass(component=self)
        self._initialize_members((index,))
        obj._index = index
        return obj

    def _setitem_when_not_present(self, index, value):
        """Perform the fundamental component item creation and storage.

        BooleanVar overrides the default implementation from IndexedComponent
        to enforce the call to _initialize_members.

        """
        obj = self._getitem_when_not_present(index)
        try:
            return obj.set_value(value)
        except:
            del self._data[index]
            raise

    def _initialize_members(self, init_set):
        """Initialize variable data for all indices in a set."""
        #
        # Initialize values
        #
        if self._value_init_rule is not None:
            #
            # Initialize values with a rule
            #
            if self.is_indexed():
                for key in init_set:
                    vardata = self._data[key]
                    val = apply_indexed_rule(
                        self, self._value_init_rule, self._parent(), key
                    )
                    val = value(val)
                    vardata.set_value(val)
            else:
                val = self._value_init_rule(self._parent())
                val = value(val)
                self.set_value(val)
        elif self._value_init_value is not None:
            #
            # Initialize values with a value
            if self._value_init_value.__class__ is dict:
                for key in init_set:
                    # Skip indices that are not in the
                    # dictionary. This arises when
                    # initializing VarList objects with a
                    # dictionary.
                    # What does this continue do here?
                    if not key in self._value_init_value:
                        continue
                    val = self._value_init_value[key]
                    vardata = self._data[key]
                    vardata.set_value(val)
            else:
                val = value(self._value_init_value)
                for key in init_set:
                    vardata = self._data[key]
                    vardata.set_value(val)

    def _pprint(self):
        """
        Print component information.
        """
        return (
            [
                ("Size", len(self)),
                ("Index", self._index_set if self.is_indexed() else None),
            ],
            self.items,
            ("Value", "Fixed", "Stale"),
            lambda k, v: [v.value, v.fixed, v.stale],
        )


class ScalarBooleanVar(BooleanVarData, BooleanVar):
    """A single variable."""

    def __init__(self, *args, **kwd):
        BooleanVarData.__init__(self, component=self)
        BooleanVar.__init__(self, *args, **kwd)
        self._index = UnindexedComponent_index

    #
    # Override abstract interface methods to first check for
    # construction
    #

    # NOTE: that we can't provide these errors for
    # fixed and stale because they are attributes

    @property
    def value(self):
        """bool : the current value of this variable."""
        if self._constructed:
            return BooleanVarData.value.fget(self)
        raise ValueError(
            "Accessing the value of variable '%s' "
            "before the Var has been constructed (there "
            "is currently no value to return)." % (self.name)
        )

    @value.setter
    def value(self, val):
        if self._constructed:
            return BooleanVarData.value.fset(self, val)
        raise ValueError(
            "Setting the value of variable '%s' "
            "before the Var has been constructed (there "
            "is currently nothing to set." % (self.name)
        )

    @property
    def domain(self):
        """BooleanSet : the domain for this variable."""
        return BooleanVarData.domain.fget(self)

    def fix(self, value=NOTSET, skip_validation=False):
        """
        Set the fixed indicator to True. Value argument is optional,
        indicating the variable should be fixed at its current value.
        """
        if self._constructed:
            return BooleanVarData.fix(self, value, skip_validation)
        raise ValueError(
            "Fixing variable '%s' "
            "before the Var has been constructed (there "
            "is currently nothing to set)." % (self.name)
        )

    def unfix(self):
        """Sets the fixed indicator to False."""
        if self._constructed:
            return BooleanVarData.unfix(self)
        raise ValueError(
            "Freeing variable '%s' "
            "before the Var has been constructed (there "
            "is currently nothing to set)." % (self.name)
        )


class SimpleBooleanVar(metaclass=RenamedClass):
    __renamed__new_class__ = ScalarBooleanVar
    __renamed__version__ = '6.0'


class IndexedBooleanVar(BooleanVar):
    """An array of variables."""

    def fix(self, value=NOTSET, skip_validation=False):
        """Fix all variables in this IndexedBooleanVar (treat as nonvariable)

        This sets the `fixed` indicator to True for every variable in
        this IndexedBooleanVar.  If ``value`` is provided, the value
        (and the ``skip_validation`` flag) are first passed to
        :py:meth:`set_value()`.

        """
        for boolean_vardata in self.values():
            boolean_vardata.fix(value, skip_validation)

    def unfix(self):
        """Unfix all variables in this IndexedBooleanVar (treat as variable)

        This sets the `fixed` indicator to False for every variable in
        this IndexedBooleanVar.

        """
        for boolean_vardata in self.values():
            boolean_vardata.unfix()

    def free(self):
        """Alias for :py:meth:`unfix`"""
        return self.unfix()

    @property
    def domain(self):
        """BooleanSet : the domain for this variable."""
        return BooleanSet

    # Because Emma wants crazy things... (Where crazy things are the ability to
    # index BooleanVars by other (integer) Vars and integer-valued
    # expressions--a thing you can do in Constraint Programming.)
    def __getitem__(self, args):
        tmp = args if args.__class__ is tuple else (args,)
        if any(
            hasattr(arg, 'is_potentially_variable') and arg.is_potentially_variable()
            for arg in tmp
        ):
            return GetItemExpression((self,) + tmp)
        return super().__getitem__(args)


@ModelComponentFactory.register("List of logical decision variables.")
class BooleanVarList(IndexedBooleanVar):
    """
    Variable-length indexed variable objects used to construct Pyomo models.
    """

    def __init__(self, **kwargs):
        self._starting_index = kwargs.pop('starting_index', 1)
        args = (Set(),)
        IndexedBooleanVar.__init__(self, *args, **kwargs)

    def construct(self, data=None):
        """Construct this component."""
        if is_debug_set(logger):
            logger.debug("Constructing variable list %s", self.name)

        # We need to ensure that the indices needed for initialization are
        # added to the underlying implicit set.  We *could* verify that the
        # indices in the initialization dict are all sequential integers,
        # OR we can just add the correct number of sequential integers and
        # then let _validate_index complain when we set the value.
        if self._value_init_value.__class__ is dict:
            for i in range(len(self._value_init_value)):
                self._index_set.add(i + self._starting_index)
        super(BooleanVarList, self).construct(data)
        # Note that the current Var initializer silently ignores
        # initialization data that is not in the underlying index set.  To
        # ensure that at least here all initialization data is added to the
        # VarList (so we get potential domain errors), we will re-set
        # everything.
        if self._value_init_value.__class__ is dict:
            for k, v in self._value_init_value.items():
                self[k] = v

    def add(self):
        """Add a variable to this list."""
        next_idx = len(self._index_set) + self._starting_index
        self._index_set.add(next_idx)
        return self[next_idx]



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import time

from pyomo.common.collections import OrderedDict

import pyomo.opt
from pyomo.opt.parallel.manager import ActionManagerError, ActionStatus, ActionHandle
from pyomo.opt.parallel.async_solver import (
    AsynchronousSolverManager,
    SolverManagerFactory,
)


@SolverManagerFactory.register("serial", doc="Synchronously execute solvers locally")
class SolverManager_Serial(AsynchronousSolverManager):
    def clear(self):
        """
        Clear manager state
        """
        super(SolverManager_Serial, self).clear()
        self.results = OrderedDict()

    def _perform_queue(self, ah, *args, **kwds):
        """
        Perform the queue operation.  This method returns the ActionHandle,
        and the ActionHandle status indicates whether the queue was successful.
        """

        opt = kwds.pop('solver', kwds.pop('opt', None))
        if opt is None:
            raise ActionManagerError(
                "No solver passed to %s, use keyword option 'solver'"
                % (type(self).__name__)
            )

        time_start = time.time()
        if isinstance(opt, str):
            with pyomo.opt.SolverFactory(opt) as _opt:
                results = _opt.solve(*args, **kwds)
        else:
            results = opt.solve(*args, **kwds)
        results.pyomo_solve_time = time.time() - time_start

        self.results[ah.id] = results
        ah.status = ActionStatus.done
        self.event_handle[ah.id].update(ah)

        return ah

    def _perform_wait_any(self):
        """
        Perform the wait_any operation.  This method returns an
        ActionHandle with the results of waiting.  If None is returned
        then the ActionManager assumes that it can call this method again.
        Note that an ActionHandle can be returned with a dummy value,
        to indicate an error.
        """
        if len(self.results) > 0:
            ah_id, result = self.results.popitem(last=False)
            self.results[ah_id] = result
            return self.event_handle[ah_id]
        return ActionHandle(
            error=True,
            explanation=(
                "No queued evaluations available in "
                "the 'serial' solver manager, which "
                "executes solvers synchronously"
            ),
        )



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
"""
Rooney Biegler model, based on Rooney, W. C. and Biegler, L. T. (2001). Design for
model parameter uncertainty using nonlinear confidence regions. AIChE Journal,
47(8), 1794-1804.
"""

from pyomo.common.dependencies import pandas as pd
import pyomo.environ as pyo


def rooney_biegler_model(data):
    """This function generates an instance of the rooney & biegler Pyomo model using 'data' as the input argument

    Parameters
    ----------
    data: pandas DataFrame, list of dictionaries, or list of json file names
        Data that is used to build an instance of the Pyomo model

    Returns
    -------
    m: an instance of the Pyomo model
        for estimating parameters and covariance
    """
    model = pyo.ConcreteModel()

    model.asymptote = pyo.Var(initialize=15)
    model.rate_constant = pyo.Var(initialize=0.5)

    def response_rule(m, h):
        expr = m.asymptote * (1 - pyo.exp(-m.rate_constant * h))
        return expr

    model.response_function = pyo.Expression(data.hour, rule=response_rule)

    def SSE_rule(m):
        return sum(
            (data.y[i] - m.response_function[data.hour[i]]) ** 2 for i in data.index
        )

    model.SSE = pyo.Objective(rule=SSE_rule, sense=pyo.minimize)

    return model


def rooney_biegler_model_opt():
    """This function generates an instance of the rooney & biegler Pyomo model

    Returns
    -------
    m: an instance of the Pyomo model
        for uncertainty propagation
    """

    model = pyo.ConcreteModel()

    model.asymptote = pyo.Var(initialize=15)
    model.rate_constant = pyo.Var(initialize=0.5)

    model.obj = pyo.Objective(
        expr=model.asymptote * (1 - pyo.exp(-model.rate_constant * 10)),
        sense=pyo.minimize,
    )
    return model



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import itertools
import pyomo.common.unittest as unittest
import pyomo.environ as pyo
from pyomo.common.dependencies import (
    scipy,
    scipy_available,
    numpy as np,
    numpy_available,
    networkx_available,
)
from pyomo.contrib.pynumero.asl import AmplInterface
from pyomo.contrib.pynumero.algorithms.solvers.implicit_functions import (
    CyIpoptSolverWrapper,
    ImplicitFunctionSolver,
    DecomposedImplicitFunctionBase,
    SccImplicitFunctionSolver,
)
from pyomo.contrib.pynumero.algorithms.solvers.cyipopt_solver import cyipopt_available

if not scipy_available or not numpy_available:
    # SciPy is only really necessary as it is a dependency of AmplInterface.
    # NumPy is directly used by the implicit function solvers.
    raise unittest.SkipTest(
        "NumPy and SciPy are needed to test the implicit function solvers"
    )
if not AmplInterface.available():
    # AmplInterface is not theoretically necessary for these solvers,
    # however it is the only AD backend implemented for PyomoNLP.
    raise unittest.SkipTest(
        "PyNumero ASL extension is necessary to test implicit function solvers"
    )


class ImplicitFunction1:
    def __init__(self):
        self._model = self._make_model()

    def _make_model(self):
        m = pyo.ConcreteModel()
        m.I = pyo.Set(initialize=[1, 2, 3])
        m.J = pyo.Set(initialize=[1, 2])
        m.x = pyo.Var(m.I, initialize=1.0)
        m.p = pyo.Var(m.J, initialize=1.0)
        # Note that this system of constraints decomposes. First con1
        # and con3 are used to solve for x[2] and x[3], then con2
        # is used to solve for x[1]
        m.con1 = pyo.Constraint(expr=m.x[2] ** 2 + m.x[3] ** 2 == m.p[1])
        m.con2 = pyo.Constraint(
            expr=2 * m.x[1] + 3 * m.x[2] - 4 * m.x[3] == m.p[1] ** 2 - m.p[2]
        )
        m.con3 = pyo.Constraint(expr=m.p[2] ** 1.5 == 2 * pyo.exp(m.x[2] / m.x[3]))
        m.obj = pyo.Objective(expr=0.0)
        return m

    def get_parameters(self):
        m = self._model
        return [m.p[1], m.p[2]]

    def get_variables(self):
        m = self._model
        return [m.x[1], m.x[2], m.x[3]]

    def get_equations(self):
        m = self._model
        return [m.con1, m.con2, m.con3]

    def get_input_output_sequence(self):
        p1_inputs = [1.0, 2.0, 3.0]
        p2_inputs = [1.0, 2.0, 3.0]
        inputs = list(itertools.product(p1_inputs, p2_inputs))

        outputs = [
            # Outputs computed by solving system with Ipopt
            (2.498253, -0.569676, 0.821869),
            (0.898530, 0.327465, 0.944863),
            (-0.589294, 0.690561, 0.723274),
            (5.033063, -0.805644, 1.162299),
            (2.977820, 0.463105, 1.336239),
            (1.080826, 0.976601, 1.022864),
            (8.327101, -0.986708, 1.423519),
            (5.922325, 0.567186, 1.636551),
            (3.711364, 1.196087, 1.252747),
        ]
        # We will iterate over these input/output pairs, set inputs,
        # solve, and check outputs. Note that these values are computed
        # with Ipopt, and the default solver for the system defining the
        # implicit function is scipy.optimize.fsolve. There is no guarantee
        # that these algorithms converge to the same solution for the
        # highly nonlinear system defining this implicit function. If some
        # of these tests start failing (e.g. because one of these algorithms
        # changes), some of these inputs may need to be omitted.
        return list(zip(inputs, outputs))


class ImplicitFunctionWithExtraVariables(ImplicitFunction1):
    """This is the same system as ImplicitFunction1, but now some
    of the hand-coded constants have been replaced by unfixed variables.
    These variables will be completely ignored and treated as constants
    by the implicit functions.

    """

    def _make_model(self):
        m = pyo.ConcreteModel()
        m.I = pyo.Set(initialize=[1, 2, 3])
        m.J = pyo.Set(initialize=[1, 2])
        m.K = pyo.Set(initialize=[1, 2, 3])
        m.x = pyo.Var(m.I, initialize=1.0)
        m.p = pyo.Var(m.J, initialize=1.0)

        # These variables will be treated as neither outputs nor
        # inputs. They are simply treated as constants.
        m.const = pyo.Var(m.K, initialize=1.0)
        m.const[1].set_value(1.0)
        m.const[2].set_value(2.0)
        m.const[3].set_value(1.5)

        m.con1 = pyo.Constraint(expr=m.const[1] * m.x[2] ** 2 + m.x[3] ** 2 == m.p[1])
        m.con2 = pyo.Constraint(
            expr=m.const[2] * m.x[1] + 3 * m.x[2] - 4 * m.x[3] == m.p[1] ** 2 - m.p[2]
        )
        m.con3 = pyo.Constraint(
            expr=m.p[2] ** m.const[3] == 2 * pyo.exp(m.x[2] / m.x[3])
        )
        m.obj = pyo.Objective(expr=0.0)
        return m


class ImplicitFunctionInputsDontAppear:
    """This is an implicit function designed to test the edge case
    where inputs do not appear in the system defining the implicit
    function (i.e. the function is constant).

    """

    def __init__(self):
        self._model = self._make_model()

    def _make_model(self):
        m = pyo.ConcreteModel()
        m.I = pyo.Set(initialize=[1, 2, 3])
        m.J = pyo.Set(initialize=[1, 2])
        m.x = pyo.Var(m.I, initialize=1.0)
        m.p = pyo.Var(m.J, initialize=1.0)
        m.con1 = pyo.Constraint(expr=m.x[2] ** 2 + m.x[3] ** 2 == 1.0)
        m.con2 = pyo.Constraint(expr=2 * m.x[1] + 3 * m.x[2] - 4 * m.x[3] == 0.0)
        m.con3 = pyo.Constraint(expr=1.0 == 2 * pyo.exp(m.x[2] / m.x[3]))
        m.obj = pyo.Objective(expr=0.0)
        return m

    def get_parameters(self):
        m = self._model
        return [m.p[1], m.p[2]]

    def get_variables(self):
        m = self._model
        return [m.x[1], m.x[2], m.x[3]]

    def get_equations(self):
        m = self._model
        return [m.con1, m.con2, m.con3]

    def get_input_output_sequence(self):
        # As the implicit function is constant, these parameter
        # values don't matter
        p1_inputs = [-1.0, 0.0]
        p2_inputs = [1.0]
        inputs = list(itertools.product(p1_inputs, p2_inputs))

        outputs = [
            # Outputs computed by solving system with Ipopt
            (2.498253, -0.569676, 0.821869),
            (2.498253, -0.569676, 0.821869),
        ]
        return list(zip(inputs, outputs))


class ImplicitFunctionNoInputs(ImplicitFunctionInputsDontAppear):
    """The same system as with inputs that don't appear, but now the
    inputs are not provided to the implicit function solver

    """

    def get_parameters(self):
        return []

    def get_input_output_sequence(self):
        inputs = [()]
        outputs = [
            # Outputs computed by solving system with Ipopt
            (2.498253, -0.569676, 0.821869)
        ]
        return list(zip(inputs, outputs))


class _TestSolver(unittest.TestCase):
    """A suite of basic tests for implicit function solvers.

    A "concrete" subclass should be defined for each implicit function
    solver. This subclass should implement get_solver_class, then
    add "test" methods that call the following methods:

        _test_implicit_function_1
        _test_implicit_function_inputs_dont_appear
        _test_implicit_function_no_inputs
        _test_implicit_function_with_extra_variables

    These methods are private so they don't get picked up on the base
    class by pytest.

    """

    def get_solver_class(self):
        raise NotImplementedError()

    def _test_implicit_function(self, ImplicitFunctionClass, **kwds):
        SolverClass = self.get_solver_class()
        fcn = ImplicitFunctionClass()
        variables = fcn.get_variables()
        parameters = fcn.get_parameters()
        equations = fcn.get_equations()

        solver = SolverClass(variables, equations, parameters, **kwds)

        for inputs, pred_outputs in fcn.get_input_output_sequence():
            solver.set_parameters(inputs)
            outputs = solver.evaluate_outputs()
            self.assertStructuredAlmostEqual(
                list(outputs), list(pred_outputs), reltol=1e-5, abstol=1e-5
            )

            solver.update_pyomo_model()
            for i, var in enumerate(variables):
                self.assertAlmostEqual(var.value, pred_outputs[i], delta=1e-5)

    def _test_implicit_function_1(self, **kwds):
        self._test_implicit_function(ImplicitFunction1, **kwds)

    def _test_implicit_function_inputs_dont_appear(self):
        self._test_implicit_function(ImplicitFunctionInputsDontAppear)

    def _test_implicit_function_no_inputs(self):
        self._test_implicit_function(ImplicitFunctionNoInputs)

    def _test_implicit_function_with_extra_variables(self):
        self._test_implicit_function(ImplicitFunctionWithExtraVariables)


class TestImplicitFunctionSolver(_TestSolver):
    def get_solver_class(self):
        return ImplicitFunctionSolver

    def test_bad_option(self):
        msg = "Option.*is invalid"
        with self.assertRaisesRegex(ValueError, msg):
            self._test_implicit_function_1(solver_options=dict(bad_option=None))

    def test_implicit_function_1(self):
        self._test_implicit_function_1()

    @unittest.skipUnless(cyipopt_available, "CyIpopt is not available")
    def test_implicit_function_1_with_cyipopt(self):
        self._test_implicit_function_1(solver_class=CyIpoptSolverWrapper)

    def test_implicit_function_inputs_dont_appear(self):
        self._test_implicit_function_inputs_dont_appear()

    def test_implicit_function_no_inputs(self):
        self._test_implicit_function_no_inputs()

    def test_implicit_function_with_extra_variables(self):
        self._test_implicit_function_with_extra_variables()


@unittest.skipUnless(networkx_available, "NetworkX is not available")
class TestSccImplicitFunctionSolver(_TestSolver):
    def get_solver_class(self):
        return SccImplicitFunctionSolver

    def test_partition_not_implemented(self):
        fcn = ImplicitFunction1()
        variables = fcn.get_variables()
        parameters = fcn.get_parameters()
        equations = fcn.get_equations()
        msg = "has not implemented"
        with self.assertRaisesRegex(NotImplementedError, msg):
            solver = DecomposedImplicitFunctionBase(variables, equations, parameters)

    def test_n_subsystems(self):
        SolverClass = self.get_solver_class()
        fcn = ImplicitFunction1()
        variables = fcn.get_variables()
        parameters = fcn.get_parameters()
        equations = fcn.get_equations()
        solver = SolverClass(variables, equations, parameters)

        # Assert that the system decomposes into two subsystems.
        self.assertEqual(solver.n_subsystems(), 2)

    def test_implicit_function_1(self):
        self._test_implicit_function_1()

    @unittest.skipUnless(cyipopt_available, "CyIpopt is not available")
    def test_implicit_function_1_with_cyipopt(self):
        self._test_implicit_function_1(solver_class=CyIpoptSolverWrapper)

    def test_implicit_function_1_no_calc_var(self):
        self._test_implicit_function_1(
            use_calc_var=False, solver_options={"maxfev": 20}
        )

    def test_implicit_function_inputs_dont_appear(self):
        self._test_implicit_function_inputs_dont_appear()

    def test_implicit_function_no_inputs(self):
        self._test_implicit_function_no_inputs()

    def test_implicit_function_with_extra_variables(self):
        self._test_implicit_function_with_extra_variables()


def _solve_with_ipopt():
    from pyomo.util.subsystems import TemporarySubsystemManager

    ipopt = pyo.SolverFactory("ipopt")
    fcn = ImplicitFunctionInputsDontAppear()
    m = fcn._model
    params = fcn.get_parameters()
    variables = fcn.get_variables()
    input_list = []
    output_list = []
    error_list = []
    for (p1, p2), _ in fcn.get_input_output_sequence():
        params[0].set_value(p1)
        params[1].set_value(p2)
        input_list.append((p1, p2))
        with TemporarySubsystemManager(to_fix=params):
            try:
                res = ipopt.solve(m, tee=True)
                pyo.assert_optimal_termination(res)
                error_list.append(False)
            except (ValueError, AssertionError, RuntimeError):
                error_list.append(True)
            output_list.append(tuple(var.value for var in variables))
    for i, (inputs, outputs) in enumerate(zip(input_list, output_list)):
        print(inputs, outputs, error_list[i])
    for outputs in output_list:
        print("(%1.6f, %1.6f, %1.6f)," % outputs)


if __name__ == "__main__":
    # unittest.main()
    _solve_with_ipopt()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________
#
# Unit Tests for pyomo.opt.problem.ampl
#

from itertools import zip_longest
import json
import os
import sys
from os.path import join
from filecmp import cmp

import pyomo.common.unittest as unittest

from pyomo.common.errors import ApplicationError
from pyomo.common.fileutils import this_file_dir
from pyomo.common.tempfiles import TempfileManager

import pyomo.opt

currdir = this_file_dir()
deleteFiles = True


class Test(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        global solvers
        import pyomo.environ

        solvers = pyomo.opt.check_available_solvers('glpk')

    def setUp(self):
        TempfileManager.push()

    def tearDown(self):
        TempfileManager.pop(remove=deleteFiles or self.currentTestPassed())

    def test3_write_nl(self):
        """Convert from AMPL to NL"""
        self.model = pyomo.opt.AmplModel(join(currdir, 'test3.mod'))
        """ Convert from MOD+DAT to NL """
        _test = TempfileManager.create_tempfile(suffix='test3.nl')
        try:
            self.model.write(_test)
        except ApplicationError:
            err = sys.exc_info()[1]
            if pyomo.common.Executable("ampl"):
                self.fail(
                    "Unexpected ApplicationError - ampl is enabled "
                    "but not available: '%s'" % str(err)
                )
            return
        except pyomo.opt.ConverterError:
            err = sys.exc_info()[1]
            if pyomo.common.Executable("ampl"):
                self.fail(
                    "Unexpected ConverterError - ampl is enabled "
                    "but not available: '%s'" % str(err)
                )
            return
        _base = join(currdir, 'test3.baseline.nl')
        with open(_test, 'r') as run, open(_base, 'r') as baseline:
            for line1, line2 in zip_longest(run, baseline):
                for _pattern in ('# problem',):
                    if line1.find(_pattern) >= 0:
                        line1 = line1[: line1.find(_pattern) + len(_pattern)]
                        line2 = line2[: line2.find(_pattern) + len(_pattern)]
                self.assertEqual(
                    line1, line2, msg="Files %s and %s differ" % (_test, _base)
                )

    def test3_write_lp(self):
        """Convert from AMPL to LP"""
        self.model = pyomo.opt.AmplModel(join(currdir, 'test3.mod'))
        _test = TempfileManager.create_tempfile(suffix='test3.lp')
        try:
            self.model.write(_test)
        except ApplicationError:
            err = sys.exc_info()[1]
            if pyomo.common.Executable("glpsol"):
                self.fail(
                    "Unexpected ApplicationError - glpsol is enabled "
                    "but not available: '%s'" % str(err)
                )
            return
        except pyomo.opt.ConverterError:
            err = sys.exc_info()[1]
            if pyomo.common.Executable("glpsol"):
                self.fail(
                    "Unexpected ConverterError - glpsol is enabled "
                    "but not available: '%s'" % str(err)
                )
            return
        _base = join(currdir, 'test3.baseline.lp')
        with open(_test, 'r') as run, open(_base, 'r') as baseline:
            for line1, line2 in zip_longest(run, baseline):
                for _pattern in ('Problem:',):
                    if line1.find(_pattern) >= 0:
                        line1 = line1[: line1.find(_pattern) + len(_pattern)]
                        line2 = line2[: line2.find(_pattern) + len(_pattern)]
                self.assertEqual(
                    line1, line2, msg="Files %s and %s differ" % (_test, _base)
                )

    def test3_write_mps(self):
        """Convert from AMPL to MPS"""
        if not pyomo.common.Executable("ampl"):
            self.skipTest("The ampl executable is not available")
        self.model = pyomo.opt.AmplModel(join(currdir, 'test3.mod'))
        _test = TempfileManager.create_tempfile(suffix='test3.mps')
        try:
            self.model.write(_test)
        except ApplicationError:
            err = sys.exc_info()[1]
            if pyomo.common.Executable("ampl"):
                self.fail(
                    "Unexpected ApplicationError - ampl is enabled "
                    "but not available: '%s'" % str(err)
                )
            return
        except pyomo.opt.ConverterError:
            err = sys.exc_info()[1]
            if pyomo.common.Executable("ampl"):
                self.fail(
                    "Unexpected ConverterError - ampl is enabled "
                    "but not available: '%s'" % str(err)
                )
            return
        _base = join(currdir, 'test3.baseline.mps')
        with open(_test, 'r') as run, open(_base, 'r') as baseline:
            for line1, line2 in zip_longest(run, baseline):
                for _pattern in ('NAME',):
                    if line1.find(_pattern) >= 0:
                        line1 = line1[: line1.find(_pattern) + len(_pattern)]
                        line2 = line2[: line2.find(_pattern) + len(_pattern)]
                self.assertEqual(
                    line1, line2, msg="Files %s and %s differ" % (_test, _base)
                )

    def test3a_write_nl(self):
        """Convert from AMPL to NL"""
        self.model = pyomo.opt.AmplModel(
            join(currdir, 'test3a.mod'), join(currdir, 'test3a.dat')
        )
        _test = TempfileManager.create_tempfile(suffix='test3a.nl')
        try:
            self.model.write(_test)
        except ApplicationError:
            err = sys.exc_info()[1]
            if pyomo.common.Executable("ampl"):
                self.fail(
                    "Unexpected ApplicationError - ampl is enabled "
                    "but not available: '%s'" % str(err)
                )
            return
        except pyomo.opt.ConverterError:
            err = sys.exc_info()[1]
            if pyomo.common.Executable("ampl"):
                self.fail(
                    "Unexpected ConverterError - ampl is enabled "
                    "but not available: '%s'" % str(err)
                )
            return
        _base = join(currdir, 'test3.baseline.nl')
        with open(_test, 'r') as run, open(_base, 'r') as baseline:
            for line1, line2 in zip_longest(run, baseline):
                for _pattern in ('# problem',):
                    if line1.find(_pattern) >= 0:
                        line1 = line1[: line1.find(_pattern) + len(_pattern)]
                        line2 = line2[: line2.find(_pattern) + len(_pattern)]
                self.assertEqual(
                    line1, line2, msg="Files %s and %s differ" % (_test, _base)
                )

    def test3a_write_lp(self):
        """Convert from AMPL to LP"""
        self.model = pyomo.opt.AmplModel(
            join(currdir, 'test3a.mod'), join(currdir, 'test3a.dat')
        )
        _test = TempfileManager.create_tempfile(suffix='test3a.lp')
        try:
            self.model.write(_test)
        except ApplicationError:
            err = sys.exc_info()[1]
            if pyomo.common.Executable("glpsol"):
                self.fail(
                    "Unexpected ApplicationError - glpsol is enabled "
                    "but not available: '%s'" % str(err)
                )
            return
        except pyomo.opt.ConverterError:
            err = sys.exc_info()[1]
            if pyomo.common.Executable("glpsol"):
                self.fail(
                    "Unexpected ConverterError - glpsol is enabled "
                    "but not available: '%s'" % str(err)
                )
            return
        _base = join(currdir, 'test3.baseline.lp')
        with open(_test, 'r') as run, open(_base, 'r') as baseline:
            for line1, line2 in zip_longest(run, baseline):
                for _pattern in ('Problem:',):
                    if line1.find(_pattern) >= 0:
                        line1 = line1[: line1.find(_pattern) + len(_pattern)]
                        line2 = line2[: line2.find(_pattern) + len(_pattern)]
                self.assertEqual(
                    line1, line2, msg="Files %s and %s differ" % (_test, _base)
                )

    def test3a_write_mps(self):
        """Convert from AMPL to MPS"""
        if not pyomo.common.Executable("ampl"):
            self.skipTest("The ampl executable is not available")
        self.model = pyomo.opt.AmplModel(
            join(currdir, 'test3a.mod'), join(currdir, 'test3a.dat')
        )
        _test = TempfileManager.create_tempfile(suffix='test3a.mps')
        try:
            self.model.write(_test)
        except ApplicationError:
            err = sys.exc_info()[1]
            if pyomo.common.Executable("ampl"):
                self.fail(
                    "Unexpected ApplicationError - ampl is enabled "
                    "but not available: '%s'" % str(err)
                )
            return
        except pyomo.opt.ConverterError:
            err = sys.exc_info()[1]
            if pyomo.common.Executable("ampl"):
                self.fail(
                    "Unexpected ConverterError - ampl is enabled "
                    "but not available: '%s'" % str(err)
                )
            return
        _base = join(currdir, 'test3.baseline.mps')
        with open(_test, 'r') as run, open(_base, 'r') as baseline:
            for line1, line2 in zip_longest(run, baseline):
                for _pattern in ('NAME',):
                    if line1.find(_pattern) >= 0:
                        line1 = line1[: line1.find(_pattern) + len(_pattern)]
                        line2 = line2[: line2.find(_pattern) + len(_pattern)]
                self.assertEqual(
                    line1, line2, msg="Files %s and %s differ" % (_test, _base)
                )

    def test3_solve(self):
        if not 'glpk' in solvers:
            self.skipTest("glpk solver is not available")
        self.model = pyomo.opt.AmplModel(join(currdir, 'test3.mod'))
        opt = pyomo.opt.SolverFactory('glpk')
        _test = TempfileManager.create_tempfile(suffix='test3.out')
        results = opt.solve(self.model, keepfiles=False)
        results.write(filename=_test, format='json')
        with (
            open(_test, 'r') as out,
            open(join(currdir, "test3.baseline.out"), 'r') as txt,
        ):
            self.assertStructuredAlmostEqual(
                json.load(txt), json.load(out), abstol=1e-6, allow_second_superset=True
            )

    def test3a_solve(self):
        if not 'glpk' in solvers:
            self.skipTest("glpk solver is not available")
        self.model = pyomo.opt.AmplModel(
            join(currdir, 'test3a.mod'), join(currdir, 'test3a.dat')
        )
        opt = pyomo.opt.SolverFactory('glpk')
        results = opt.solve(self.model, keepfiles=False)
        _test = TempfileManager.create_tempfile(suffix='test3a.out')
        results.write(filename=_test, format='json')
        with (
            open(_test, 'r') as out,
            open(join(currdir, "test3.baseline.out"), 'r') as txt,
        ):
            self.assertStructuredAlmostEqual(
                json.load(txt), json.load(out), abstol=1e-6, allow_second_superset=True
            )


if __name__ == "__main__":
    deleteFiles = False
    unittest.main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

"""Tests the equality set propagation module."""

import pyomo.common.unittest as unittest
from pyomo.common.errors import InfeasibleConstraintException

from pyomo.environ import (
    ConcreteModel,
    Constraint,
    RangeSet,
    TransformationFactory,
    Var,
    value,
)


class TestEqualityPropagate(unittest.TestCase):
    """Tests equality set variable attribute propagation."""

    def test_fixed_var_propagate(self):
        """Test for transitivity in a variable equality set."""
        m = ConcreteModel()
        m.v1 = Var(initialize=1)
        m.v2 = Var(initialize=2)
        m.v3 = Var(initialize=3)
        m.v4 = Var(initialize=4)
        m.c1 = Constraint(expr=m.v1 == m.v2)
        m.c2 = Constraint(expr=m.v2 == m.v3)
        m.c3 = Constraint(expr=m.v3 == m.v4)
        m.v2.fix()

        m.s = RangeSet(4)
        m.x = Var(m.s, initialize=5)
        m.c = Constraint(m.s)
        m.c.add(1, expr=m.x[1] == m.x[3])
        m.c.add(2, expr=m.x[2] == m.x[4])
        m.c.add(3, expr=m.x[2] == m.x[3])
        m.c.add(4, expr=m.x[1] == 1)

        m.y = Var([1, 2], initialize=3)
        m.c_too = Constraint(expr=m.y[1] == m.y[2])

        m.z1 = Var()
        m.z2 = Var()
        m.ignore_me = Constraint(expr=m.y[1] + m.z1 + m.z2 <= 0)

        TransformationFactory('contrib.propagate_fixed_vars').apply_to(m)
        self.assertTrue(m.v1.fixed)
        self.assertTrue(m.v2.fixed)
        self.assertTrue(m.v3.fixed)
        self.assertTrue(m.v4.fixed)
        self.assertEqual(value(m.v4), 2)

        self.assertTrue(m.x[1].fixed)
        self.assertTrue(m.x[2].fixed)
        self.assertTrue(m.x[3].fixed)
        self.assertTrue(m.x[4].fixed)
        self.assertEqual(value(m.x[4]), 1)

        self.assertFalse(m.y[1].fixed)
        self.assertFalse(m.y[2].fixed)
        self.assertFalse(m.z1.fixed)
        self.assertFalse(m.z2.fixed)
        # m.display()

    def test_fixed_var_propagate_backwards(self):
        """Test backwards propagation through equality set."""
        m = ConcreteModel()
        m.v1 = Var(initialize=1)
        m.v2 = Var(initialize=2)
        m.v3 = Var(initialize=3)
        m.v4 = Var(initialize=4)
        m.c1 = Constraint(expr=m.v1 == m.v2)
        m.c2 = Constraint(expr=m.v2 == m.v3)
        m.c3 = Constraint(expr=m.v3 == m.v4)
        m.v4.fix()
        # check to make sure that all the v's have the same equality set. John
        # had found a logic error.
        TransformationFactory('contrib.propagate_fixed_vars').apply_to(m)
        # m.display()
        self.assertTrue(m.v1.fixed)
        self.assertTrue(m.v2.fixed)
        self.assertTrue(m.v3.fixed)
        self.assertTrue(m.v4.fixed)
        self.assertEqual(value(m.v4), 4)

    def test_var_fix_revert(self):
        """Test to make sure that variable fixing reversion works."""
        m = ConcreteModel()
        m.v1 = Var(initialize=1)
        m.v2 = Var(initialize=2)
        m.v3 = Var(initialize=3)
        m.v4 = Var(initialize=4)
        m.c1 = Constraint(expr=m.v1 == m.v2)
        m.c2 = Constraint(expr=m.v2 == m.v3)
        m.c3 = Constraint(expr=m.v3 == m.v4)
        m.v2.fix()
        fvp = TransformationFactory('contrib.propagate_fixed_vars')
        fvp.apply_to(m, tmp=True)
        self.assertTrue(m.v1.fixed)
        self.assertTrue(m.v2.fixed)
        self.assertTrue(m.v3.fixed)
        self.assertTrue(m.v4.fixed)
        fvp.revert(m)
        self.assertFalse(m.v1.fixed)
        self.assertTrue(m.v2.fixed)
        self.assertFalse(m.v3.fixed)
        self.assertFalse(m.v4.fixed)

    def test_var_fix_accounts_for_constants(self):
        """Test to make sure that constraints of the form x == y + constant
        are handled correctly when propagating fixed variables."""
        m = ConcreteModel()
        m.v = Var(initialize=1.0)
        m.v2 = Var(initialize=1.0)
        m.v3 = Var(initialize=1.0)
        m.c = Constraint(expr=m.v - m.v2 + m.v3 == 0)
        m.v.fix()
        m.v4 = Var(initialize=1.0)
        m.c2 = Constraint(expr=m.v2 == m.v4)
        m.v4.fix()
        TransformationFactory('contrib.propagate_fixed_vars').apply_to(m)

        self.assertTrue(m.v.fixed)
        self.assertEqual(value(m.v), 1)
        self.assertTrue(m.v4.fixed)
        self.assertEqual(value(m.v4), 1)
        self.assertTrue(m.v2.fixed)
        self.assertEqual(value(m.v2), 1)
        # v3 doesn't get fixed at all since m.c is not in the form x == y.
        self.assertFalse(m.v3.fixed)

    def test_var_bound_propagate(self):
        """Test for transitivity in a variable equality set."""
        m = ConcreteModel()
        m.v1 = Var(initialize=1, bounds=(1, 3))
        m.v2 = Var(initialize=2, bounds=(0, 8))
        m.v3 = Var(initialize=3, bounds=(2, 4))
        m.v4 = Var(initialize=4, bounds=(0, 5))
        m.c1 = Constraint(expr=m.v1 == m.v2)
        m.c2 = Constraint(expr=m.v2 == m.v3)
        m.c3 = Constraint(expr=m.v3 == m.v4)

        m.s = RangeSet(4)
        m.x = Var(m.s, initialize=5)
        m.x[2].setlb(-1)
        m.c = Constraint(m.s)
        m.c.add(1, expr=m.x[1] == m.x[3])
        m.c.add(2, expr=m.x[2] == m.x[4])
        m.c.add(3, expr=m.x[2] == m.x[3])
        m.c.add(4, expr=m.x[1] == 1)

        m.y = Var([1, 2], initialize=3)
        m.y[1].setub(3)
        m.y[2].setub(15)
        m.c_too = Constraint(expr=m.y[1] == m.y[2])

        m.z1 = Var(bounds=(1, 2))
        m.z2 = Var(bounds=(3, 4))
        m.ignore_me = Constraint(expr=m.y[1] + m.z1 + m.z2 <= 0)

        TransformationFactory('contrib.propagate_eq_var_bounds').apply_to(m)

        self.assertEqual(value(m.v1.lb), 2)
        self.assertEqual(value(m.v1.lb), value(m.v2.lb))
        self.assertEqual(value(m.v1.lb), value(m.v3.lb))
        self.assertEqual(value(m.v1.lb), value(m.v4.lb))
        self.assertEqual(value(m.v1.ub), 3)
        self.assertEqual(value(m.v1.ub), value(m.v2.ub))
        self.assertEqual(value(m.v1.ub), value(m.v3.ub))
        self.assertEqual(value(m.v1.ub), value(m.v4.ub))

        for i in m.s:
            self.assertEqual(value(m.x[i].lb), -1)

        self.assertEqual(value(m.y[1].ub), 3)
        self.assertEqual(value(m.y[2].ub), 3)
        self.assertEqual(value(m.y[1].lb), None)
        self.assertEqual(value(m.y[1].lb), None)

        self.assertEqual(value(m.z1.ub), 2)
        self.assertEqual(value(m.z2.ub), 4)
        self.assertEqual(value(m.z1.lb), 1)
        self.assertEqual(value(m.z2.lb), 3)

    def test_var_bound_propagate_crossover(self):
        """Test for error message when variable bound crosses over."""
        m = ConcreteModel()
        m.v1 = Var(initialize=1, bounds=(1, 3))
        m.v2 = Var(initialize=5, bounds=(4, 8))
        m.c1 = Constraint(expr=m.v1 == m.v2)
        xfrm = TransformationFactory('contrib.propagate_eq_var_bounds')
        with self.assertRaisesRegex(
            InfeasibleConstraintException,
            "Variable v2 has a lower bound 4 > the "
            "upper bound 3 of variable v1, but they "
            "are linked by equality constraints",
        ):
            xfrm.apply_to(m)

    def test_var_bound_propagate_revert(self):
        """Test to make sure bound propagation revert works."""
        m = ConcreteModel()
        m.v1 = Var(initialize=1, bounds=(1, 3))
        m.v2 = Var(initialize=2, bounds=(0, 8))
        m.v3 = Var(initialize=3, bounds=(2, 4))
        m.v4 = Var(initialize=4, bounds=(0, 5))
        m.c1 = Constraint(expr=m.v1 == m.v2)
        m.c2 = Constraint(expr=m.v2 == m.v3)
        m.c3 = Constraint(expr=m.v3 == m.v4)
        xfrm = TransformationFactory('contrib.propagate_eq_var_bounds')
        xfrm.apply_to(m, tmp=True)
        self.assertEqual(value(m.v1.lb), 2)
        self.assertEqual(value(m.v1.lb), value(m.v2.lb))
        self.assertEqual(value(m.v1.lb), value(m.v3.lb))
        self.assertEqual(value(m.v1.lb), value(m.v4.lb))
        self.assertEqual(value(m.v1.ub), 3)
        self.assertEqual(value(m.v1.ub), value(m.v2.ub))
        self.assertEqual(value(m.v1.ub), value(m.v3.ub))
        self.assertEqual(value(m.v1.ub), value(m.v4.ub))
        xfrm.revert(m)
        self.assertEqual(value(m.v1.lb), 1)
        self.assertEqual(value(m.v2.lb), 0)
        self.assertEqual(value(m.v3.lb), 2)
        self.assertEqual(value(m.v4.lb), 0)
        self.assertEqual(value(m.v1.ub), 3)
        self.assertEqual(value(m.v2.ub), 8)
        self.assertEqual(value(m.v3.ub), 4)
        self.assertEqual(value(m.v4.ub), 5)


if __name__ == '__main__':
    unittest.main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from abc import abstractmethod
from collections.abc import Mapping, Sequence
import datetime
import time
from io import StringIO

from pyomo.common.collections import ComponentMap
from pyomo.common.errors import ApplicationError, DeveloperError, PyomoException
from pyomo.common.numeric_types import value
from pyomo.common.tee import TeeStream, capture_output
from pyomo.common.timing import HierarchicalTimer
from pyomo.contrib.solver.common.base import SolverBase
from pyomo.contrib.solver.common.results import (
    Results,
    SolutionStatus,
    TerminationCondition,
)
from pyomo.contrib.solver.common.util import (
    IncompatibleModelError,
    NoDualsError,
    NoOptimalSolutionError,
    NoReducedCostsError,
    NoSolutionError,
)
from pyomo.contrib.solver.solvers.knitro.config import KnitroConfig
from pyomo.contrib.solver.solvers.knitro.engine import Engine
from pyomo.contrib.solver.solvers.knitro.package import PackageChecker
from pyomo.contrib.solver.solvers.knitro.solution import (
    SolutionLoader,
    SolutionProvider,
)
from pyomo.contrib.solver.solvers.knitro.typing import ItemData, ItemType, ValueType
from pyomo.contrib.solver.solvers.knitro.utils import KnitroModelData
from pyomo.core.base.block import BlockData
from pyomo.core.base.constraint import ConstraintData
from pyomo.core.base.var import VarData
from pyomo.core.staleflag import StaleFlagManager


class KnitroSolverBase(SolutionProvider, PackageChecker, SolverBase):
    CONFIG = KnitroConfig()
    config: KnitroConfig

    _engine: Engine
    _model_data: KnitroModelData
    _stream: StringIO
    _saved_var_values: dict[int, float | None]

    def __init__(self, **kwds) -> None:
        PackageChecker.__init__(self)
        SolverBase.__init__(self, **kwds)
        self._engine = Engine()
        self._model_data = KnitroModelData()
        self._stream = StringIO()
        self._saved_var_values = {}

    def solve(self, model: BlockData, **kwds) -> Results:
        start_timestamp = datetime.datetime.now(datetime.timezone.utc)
        tick = time.perf_counter()
        self._check_available()

        config = self._build_config(**kwds)
        timer = config.timer or HierarchicalTimer()

        StaleFlagManager.mark_all_as_stale()

        self._presolve(model, config, timer)
        self._validate_problem()

        if config.restore_variable_values_after_solve:
            self._save_var_values()

        self._stream = StringIO()
        with capture_output(TeeStream(self._stream, *config.tee), capture_fd=True):
            self._solve(config, timer)

        if config.restore_variable_values_after_solve:
            self._restore_var_values()

        results = self._postsolve(config, timer)

        tock = time.perf_counter()

        results.timing_info.start_timestamp = start_timestamp
        results.timing_info.wall_time = tock - tick
        return results

    def _build_config(self, **kwds) -> KnitroConfig:
        return self.config(value=kwds, preserve_implicit=True)  # type: ignore

    def _validate_problem(self) -> None:
        if len(self._model_data.objs) > 1:
            msg = f"{self.name} does not support multiple objectives."
            raise IncompatibleModelError(msg)

    def _check_available(self) -> None:
        avail = self.available()
        if not avail:
            msg = f"Solver {self.name} is not available: {avail}."
            raise ApplicationError(msg)

    def _save_var_values(self) -> None:
        self._saved_var_values.clear()
        for var in self._get_vars():
            self._saved_var_values[id(var)] = value(var.value)

    def _restore_var_values(self) -> None:
        StaleFlagManager.mark_all_as_stale(delayed=True)
        for var in self._get_vars():
            var.set_value(self._saved_var_values[id(var)])
        StaleFlagManager.mark_all_as_stale()

    @abstractmethod
    def _presolve(
        self, model: BlockData, config: KnitroConfig, timer: HierarchicalTimer
    ) -> None:
        raise NotImplementedError

    @abstractmethod
    def _solve(self, config: KnitroConfig, timer: HierarchicalTimer) -> None:
        raise NotImplementedError

    def _postsolve(self, config: KnitroConfig, timer: HierarchicalTimer) -> Results:
        results = Results()
        results.solver_name = self.name
        results.solver_version = self.version()
        results.solver_log = self._stream.getvalue()
        results.solver_config = config
        results.solution_status = self._engine.get_solution_status()
        results.termination_condition = self._engine.get_termination_condition()
        results.incumbent_objective = self._engine.get_obj_value()

        if self._is_mip():
            results.objective_bound = self._engine.get_obj_bound()

        if self._is_mip():
            results.extra_info.mip_number_nodes = self._engine.get_mip_number_nodes()
            results.extra_info.mip_abs_gap = self._engine.get_mip_abs_gap()
            results.extra_info.mip_rel_gap = self._engine.get_mip_rel_gap()
            results.extra_info.mip_number_solves = self._engine.get_mip_number_solves()
        else:
            results.extra_info.number_iters = self._engine.get_number_iters()
        results.timing_info.solve_time = self._engine.get_solve_time()
        results.timing_info.timer = timer

        if (
            config.raise_exception_on_nonoptimal_result
            and results.termination_condition
            != TerminationCondition.convergenceCriteriaSatisfied
        ):
            raise NoOptimalSolutionError()

        results.solution_loader = SolutionLoader(
            self,
            has_primals=results.solution_status
            not in {SolutionStatus.infeasible, SolutionStatus.noSolution},
            has_reduced_costs=results.solution_status == SolutionStatus.optimal,
            has_duals=results.solution_status
            not in {SolutionStatus.infeasible, SolutionStatus.noSolution},
        )
        if config.load_solutions:
            timer.start("load_solutions")
            results.solution_loader.load_vars()
            timer.stop("load_solutions")

        return results

    def get_values(
        self,
        item_type: type[ItemType],
        value_type: ValueType,
        items: Sequence[ItemType] | None = None,
        *,
        exists: bool,
        solution_id: int | None = None,
    ) -> Mapping[ItemType, float]:
        error_type = self._get_error_type(item_type, value_type)
        if not exists:
            raise error_type()
        # KNITRO only supports a single solution
        assert solution_id is None
        if items is None:
            items = self._get_items(item_type)
        x = self._engine.get_values(item_type, value_type, items)
        if x is None:
            raise error_type()
        sign = value_type.sign
        return ComponentMap([(k, sign * xk) for k, xk in zip(items, x)])

    def get_num_solutions(self) -> int:
        return self._engine.get_num_solutions()

    def _get_vars(self) -> list[VarData]:
        return self._model_data.variables

    def _get_items(self, item_type: type[ItemType]) -> Sequence[ItemType]:
        maps = {
            VarData: self._model_data.variables,
            ConstraintData: self._model_data.cons,
        }
        return maps[item_type]

    @staticmethod
    def _get_error_type(
        item_type: type[ItemData], value_type: ValueType
    ) -> type[PyomoException]:
        if item_type is VarData and value_type == ValueType.PRIMAL:
            return NoSolutionError
        elif item_type is VarData and value_type == ValueType.DUAL:
            return NoReducedCostsError
        elif item_type is ConstraintData and value_type == ValueType.DUAL:
            return NoDualsError
        raise DeveloperError(
            f"Unsupported KNITRO item type {item_type} and value type {value_type}."
        )

    def _is_mip(self) -> bool:
        for var in self._model_data.variables:
            if var.is_integer() or var.is_binary():
                return True
        return False



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from collections.abc import Iterable, Mapping, MutableSet, Sequence

from pyomo.common.collections import ComponentMap, ComponentSet
from pyomo.common.numeric_types import value
from pyomo.contrib.solver.common.util import collect_vars_and_named_exprs
from pyomo.contrib.solver.solvers.knitro.typing import Function
from pyomo.core.base.block import BlockData
from pyomo.core.base.constraint import Constraint, ConstraintData
from pyomo.core.base.enums import SortComponents
from pyomo.core.base.expression import Expression
from pyomo.core.base.objective import Objective, ObjectiveData
from pyomo.core.base.var import VarData
from pyomo.core.expr.calculus.diff_with_pyomo import reverse_sd


def get_active_objectives(block: BlockData) -> list[ObjectiveData]:
    """Retrieve all active ObjectiveData objects from a Pyomo Block.

    Args:
        block (BlockData): The Pyomo block to search for objectives.

    Returns:
        list[ObjectiveData]: A sorted list of all active objectives in the block.

    """
    generator = block.component_data_objects(
        Objective, descend_into=True, active=True, sort=True
    )
    return list(generator)


def get_active_constraints(block: BlockData) -> list[ConstraintData]:
    """Retrieve all active ConstraintData objects from a Pyomo Block.

    Args:
        block (BlockData): The Pyomo block to search for constraints.

    Returns:
        list[ConstraintData]: A sorted list of all active constraints in the block.

    """
    generator = block.component_data_objects(
        Constraint, descend_into=True, active=True, sort=SortComponents.deterministic
    )
    return list(generator)


class KnitroModelData:
    """Intermediate representation of a Pyomo model for KNITRO.

    Collects all active objectives, constraints, and referenced variables from a Pyomo Block.
    This class is used to extract and organize model data before passing it to the solver.

    Attributes:
        objs (list[ObjectiveData]): list of active objectives.
        cons (list[ConstraintData]): list of active constraints.
        variables (list[VarData]): list of all referenced variables.

    """

    objs: list[ObjectiveData]
    cons: list[ConstraintData]
    variables: list[VarData]
    _vars: MutableSet[VarData]

    def __init__(self, block: BlockData | None = None) -> None:
        """Initialize a Problem instance.

        Args:
            block (BlockData | None): Pyomo block to initialize from. If None,
                creates an empty problem that can be populated later.

        """
        self._vars = ComponentSet()
        self.objs = []
        self.cons = []
        self.variables = []
        if block is not None:
            self.add_block(block)

    def clear(self) -> None:
        """Clear all objectives, constraints, and variables from the problem."""
        self.objs.clear()
        self.cons.clear()
        self.variables.clear()
        self._vars.clear()

    def set_block(self, block: BlockData) -> None:
        """Replace the current problem data with data from a new block.

        Args:
            block (BlockData): The Pyomo block to extract data from.

        """
        self.clear()
        self.add_block(block)

    def add_block(self, block: BlockData) -> None:
        """Add objectives, constraints, and variables from a block to the problem.

        Args:
            block (BlockData): The Pyomo block to extract data from.

        """
        new_objs = get_active_objectives(block)
        new_cons = get_active_constraints(block)
        self.objs.extend(new_objs)
        self.cons.extend(new_cons)

        # Collect variables from objectives
        for obj in new_objs:
            _, variables, _, _ = collect_vars_and_named_exprs(obj.expr)  # type: ignore
            for var in variables:
                self._vars.add(var)

        # Collect variables from constraints
        for con in new_cons:
            _, variables, _, _ = collect_vars_and_named_exprs(con.body)  # type: ignore
            for var in variables:
                self._vars.add(var)

        # Update the variables list with unique variables only
        self.variables = list(self._vars)


def set_var_values(
    variables: Iterable[VarData], values: Sequence[float], var_map: Mapping[int, int]
) -> None:
    """Set the values of a list of Pyomo variables from a list of values.

    Args:
        variables (Iterable[VarData]): The variables to set.
        values (list[float]): The list of values to assign to the variables.
        var_map (Mapping[int, int]): A mapping from variable id to index in the
            values list.

    """
    for var in variables:
        i = var_map[id(var)]
        var.set_value(values[i])


class NonlinearExpressionData(Function):
    """Holds the data required to evaluate a non-linear expression.

    This class stores a Pyomo expression along with its variables and can compute
    gradient and Hessian information for use with optimization solvers.

    Attributes:
        variables (list[VarData]): list of variables referenced in the expression.
        func_expr (Expression): The Pyomo expression representing the non-linear function.
        grad_map (Mapping[VarData, Expression]): Gradient expressions mapped by variable.
        hess_map (Mapping[tuple[VarData, VarData], Expression]): Hessian expressions
            mapped by variable pairs.
        diff_order (int): Level of differentiation to compute:
            - 0: function evaluation only
            - 1: function + gradient
            - 2: function + gradient + hessian

    """

    variables: list[VarData]
    func_expr: Expression
    grad_map: Mapping[VarData, Expression]
    hess_map: Mapping[tuple[VarData, VarData], Expression]
    diff_order: int

    def __init__(
        self,
        expr: Expression,
        variables: Iterable[VarData],
        var_map: Mapping[int, int],
        diff_order: int = 0,
    ) -> None:
        """Initialize NonlinearExpressionData.

        Args:
            expr (Expression): The Pyomo expression to evaluate.
            variables (Iterable[VarData]): Variables referenced in the expression.
            diff_order (int): Level of differentiation to compute:
                - 0: function evaluation only
                - 1: function + gradient
                - 2: function + gradient + hessian

        """
        self.func_expr = expr
        self.variables = list(variables)
        self.diff_order = diff_order
        self._var_map = var_map
        if diff_order >= 1:
            self.compute_gradient()
        if diff_order >= 2:
            self.compute_hessian()

    @property
    def grad_vars(self) -> list[VarData]:
        """Get the list of variables for which gradients are available.

        Returns:
            list[VarData]: Variables with gradient information.

        """
        return list(self.grad_map.keys())

    @property
    def hess_vars(self) -> list[tuple[VarData, VarData]]:
        """Get the list of variable pairs for which Hessian entries are available.

        Returns:
            list[tuple[VarData, VarData]]: Variable pairs with Hessian information.

        """
        return list(self.hess_map.keys())

    def compute_gradient(self) -> None:
        """Compute gradient expressions for the nonlinear expression.

        This method computes the gradient of the expression with respect to all
        variables and stores the results in the grad attribute.
        """
        derivative = reverse_sd(self.func_expr)
        variables = ComponentSet(self.variables)
        self.grad_map = ComponentMap()
        for v, expr in derivative.items():
            if v in variables:
                self.grad_map[v] = expr

    def compute_hessian(self) -> None:
        """Compute Hessian expressions for the nonlinear expression.

        This method computes the Hessian matrix of the expression with respect to all
        variables and stores the results in the hess attribute. Only the upper triangle
        of the Hessian is stored to avoid redundancy.

        Note:
            This method requires that compute_gradient() has been called first.

        """
        variables = ComponentSet(self.variables)
        self.hess_map = ComponentMap()
        for v1, grad_expr in self.grad_map.items():
            derivative = reverse_sd(grad_expr)
            for v2, hess_expr in derivative.items():
                if v2 not in variables:
                    continue
                # Store only upper triangle: ensure var1 <= var2 by ID
                var1, var2 = (v1, v2) if id(v1) <= id(v2) else (v2, v1)
                if (var1, var2) not in self.hess_map:
                    self.hess_map[(var1, var2)] = hess_expr

    def evaluate(self, x: list[float]) -> float:
        set_var_values(self.variables, x, self._var_map)
        return value(self.func_expr)

    def gradient(self, x: list[float]) -> list[float]:
        set_var_values(self.variables, x, self._var_map)
        return [value(g) for g in self.grad_map.values()]

    def hessian(self, x: list[float], mu: float) -> list[float]:
        set_var_values(self.variables, x, self._var_map)
        return [mu * value(h) for h in self.hess_map.values()]



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import pyomo.common.unittest as unittest
from pyomo.common.errors import MouseTrap, DeveloperError
from pyomo.common.log import LoggingIntercept
import logging

from pyomo.core.expr import InequalityExpression
from pyomo.core.expr.sympy_tools import sympy_available
from pyomo.core.plugins.transform.logical_to_linear import (
    update_boolean_vars_from_binary,
)
from pyomo.environ import (
    ConcreteModel,
    BooleanConstant,
    BooleanVar,
    LogicalConstraint,
    TransformationFactory,
    RangeSet,
    Var,
    Constraint,
    ExternalFunction,
    ComponentMap,
    value,
    BooleanSet,
    land,
    lor,
    atleast,
    atmost,
    exactly,
    Block,
    Binary,
    LogicalConstraintList,
)
from pyomo.gdp import Disjunct, Disjunction
from pyomo.repn import generate_standard_repn
from io import StringIO


def _generate_boolean_model(nvars):
    m = ConcreteModel()
    m.s = RangeSet(nvars)
    m.Y = BooleanVar(m.s)
    # make sure all the variables are used in at least one logical constraint
    m.constraint = LogicalConstraint(expr=exactly(2, m.Y))
    return m


def _constrs_contained_within(test_case, test_constr_tuples, constraint_list):
    """Checks to see if constraints defined by test_constr_tuples are in the
    constraint list.

    Parameters
    ----------
    constraint_list : Constraint
    test_constr_tuples : list of tuple
    test_case : unittest.TestCase

    """

    # Move const term from body
    def _move_const_from_body(lower, repn, upper):
        if repn.constant is not None and not repn.constant == 0:
            if lower is not None:
                lower -= repn.constant
            if upper is not None:
                upper -= repn.constant
        return value(lower), repn, value(upper)

    def _repns_match(repn, test_repn):
        if not len(repn.linear_vars) == len(test_repn.linear_vars):
            return False
        coef_map = ComponentMap(
            (var, coef) for var, coef in zip(repn.linear_vars, repn.linear_coefs)
        )
        for var, coef in zip(test_repn.linear_vars, test_repn.linear_coefs):
            if not coef_map.get(var, 0) == coef:
                return False
        return True

    constr_list_tuples = [
        _move_const_from_body(
            constr.lower, generate_standard_repn(constr.body), constr.upper
        )
        for constr in constraint_list.values()
    ]
    for test_lower, test_body, test_upper in test_constr_tuples:
        test_repn = generate_standard_repn(test_body)
        test_lower, test_repn, test_upper = _move_const_from_body(
            test_lower, test_repn, test_upper
        )
        found_match = False
        # Make sure one of the list tuples matches
        for lower, repn, upper in constr_list_tuples:
            if (
                lower == test_lower
                and upper == test_upper
                and _repns_match(repn, test_repn)
            ):
                found_match = True
                break
        test_case.assertTrue(
            found_match,
            "{} <= {} <= {} was not found in constraint list.".format(
                test_lower, test_body, test_upper
            ),
        )


@unittest.skipUnless(sympy_available, "Sympy not available")
class TestAtomicTransformations(unittest.TestCase):
    def test_implies(self):
        m = ConcreteModel()
        m.x = BooleanVar()
        m.y = BooleanVar()
        m.p = LogicalConstraint(expr=m.x.implies(m.y))
        TransformationFactory('core.logical_to_linear').apply_to(m)
        _constrs_contained_within(
            self,
            [
                (
                    1,
                    (1 - m.x.get_associated_binary()) + m.y.get_associated_binary(),
                    None,
                )
            ],
            m.logic_to_linear.transformed_constraints,
        )

    def test_literal(self):
        m = ConcreteModel()
        m.Y = BooleanVar()
        m.p = LogicalConstraint(expr=m.Y)
        TransformationFactory('core.logical_to_linear').apply_to(m)
        _constrs_contained_within(
            self,
            [(1, m.Y.get_associated_binary(), 1)],
            m.logic_to_linear.transformed_constraints,
        )

    def test_constant_True(self):
        m = ConcreteModel()
        m.p = LogicalConstraint(expr=True)
        self.assertExpressionsStructurallyEqual(m.p.expr, BooleanConstant(True))
        TransformationFactory('core.logical_to_linear').apply_to(m)
        self.assertIsNotNone(m.component('logic_to_linear'))
        self.assertEqual(0, len(m.logic_to_linear.augmented_vars_asbinary))
        self.assertEqual(0, len(m.logic_to_linear.transformed_constraints))
        self.assertEqual(0, len(m.logic_to_linear.augmented_vars))

    def test_constant_False(self):
        m = ConcreteModel()
        m.p = LogicalConstraint(expr=False)
        self.assertExpressionsStructurallyEqual(m.p.expr, BooleanConstant(False))
        TransformationFactory('core.logical_to_linear').apply_to(m)
        self.assertIsNotNone(m.component('logic_to_linear'))
        self.assertEqual(0, len(m.logic_to_linear.augmented_vars_asbinary))
        self.assertEqual(1, len(m.logic_to_linear.transformed_constraints))
        self.assertExpressionsStructurallyEqual(
            m.logic_to_linear.transformed_constraints[1].expr,
            InequalityExpression((1, 0), False),
        )
        self.assertEqual(0, len(m.logic_to_linear.augmented_vars))

    def test_nothing_to_do(self):
        m = ConcreteModel()
        m.p = LogicalConstraint()
        TransformationFactory('core.logical_to_linear').apply_to(m)
        self.assertIsNone(m.component('logic_to_linear'))
        self.assertFalse(m.p.active)

    def test_deactivate_empty_logical_constraint_container(self):
        m = ConcreteModel()
        m.propositions = LogicalConstraintList()
        TransformationFactory('core.logical_to_linear').apply_to(m)
        self.assertIsNone(m.component('logic_to_linear'))
        self.assertFalse(m.propositions.active)


@unittest.skipUnless(sympy_available, "Sympy not available")
class TestLogicalToLinearTransformation(unittest.TestCase):
    def test_longer_statement(self):
        m = ConcreteModel()
        m.s = RangeSet(3)
        m.Y = BooleanVar(m.s)
        m.p = LogicalConstraint(expr=m.Y[1].implies(lor(m.Y[2], m.Y[3])))
        TransformationFactory('core.logical_to_linear').apply_to(m)
        _constrs_contained_within(
            self,
            [
                (
                    1,
                    m.Y[2].get_associated_binary()
                    + m.Y[3].get_associated_binary()
                    + (1 - m.Y[1].get_associated_binary()),
                    None,
                )
            ],
            m.logic_to_linear.transformed_constraints,
        )

    def test_xfrm_atleast_statement(self):
        m = ConcreteModel()
        m.s = RangeSet(3)
        m.Y = BooleanVar(m.s)
        m.p = LogicalConstraint(expr=atleast(2, m.Y[1], m.Y[2], m.Y[3]))
        TransformationFactory('core.logical_to_linear').apply_to(m)
        _constrs_contained_within(
            self,
            [
                (
                    2,
                    m.Y[1].get_associated_binary()
                    + m.Y[2].get_associated_binary()
                    + m.Y[3].get_associated_binary(),
                    None,
                )
            ],
            m.logic_to_linear.transformed_constraints,
        )

    def test_xfrm_atmost_statement(self):
        m = ConcreteModel()
        m.s = RangeSet(3)
        m.Y = BooleanVar(m.s)
        m.p = LogicalConstraint(expr=atmost(2, m.Y[1], m.Y[2], m.Y[3]))
        TransformationFactory('core.logical_to_linear').apply_to(m)
        _constrs_contained_within(
            self,
            [
                (
                    None,
                    m.Y[1].get_associated_binary()
                    + m.Y[2].get_associated_binary()
                    + m.Y[3].get_associated_binary(),
                    2,
                )
            ],
            m.logic_to_linear.transformed_constraints,
        )

    def test_xfrm_exactly_statement(self):
        m = ConcreteModel()
        m.s = RangeSet(3)
        m.Y = BooleanVar(m.s)
        m.p = LogicalConstraint(expr=exactly(2, m.Y[1], m.Y[2], m.Y[3]))
        TransformationFactory('core.logical_to_linear').apply_to(m)
        _constrs_contained_within(
            self,
            [
                (
                    2,
                    m.Y[1].get_associated_binary()
                    + m.Y[2].get_associated_binary()
                    + m.Y[3].get_associated_binary(),
                    2,
                )
            ],
            m.logic_to_linear.transformed_constraints,
        )

    def test_xfrm_special_atoms_nonroot(self):
        m = ConcreteModel()
        m.s = RangeSet(3)
        m.Y = BooleanVar(m.s)
        m.p = LogicalConstraint(expr=m.Y[1].implies(atleast(2, m.Y[1], m.Y[2], m.Y[3])))
        TransformationFactory('core.logical_to_linear').apply_to(m)
        Y_aug = m.logic_to_linear.augmented_vars
        self.assertEqual(len(Y_aug), 1)
        self.assertEqual(Y_aug[1].domain, BooleanSet)
        _constrs_contained_within(
            self,
            [
                (
                    None,
                    sum(m.Y[:].get_associated_binary())
                    - (1 + 2 * Y_aug[1].get_associated_binary()),
                    0,
                ),
                (
                    1,
                    (1 - m.Y[1].get_associated_binary())
                    + Y_aug[1].get_associated_binary(),
                    None,
                ),
                (
                    None,
                    2
                    - 2 * (1 - Y_aug[1].get_associated_binary())
                    - sum(m.Y[:].get_associated_binary()),
                    0,
                ),
            ],
            m.logic_to_linear.transformed_constraints,
        )

        m = ConcreteModel()
        m.s = RangeSet(3)
        m.Y = BooleanVar(m.s)
        m.p = LogicalConstraint(expr=m.Y[1].implies(atmost(2, m.Y[1], m.Y[2], m.Y[3])))
        TransformationFactory('core.logical_to_linear').apply_to(m)
        Y_aug = m.logic_to_linear.augmented_vars
        self.assertEqual(len(Y_aug), 1)
        self.assertEqual(Y_aug[1].domain, BooleanSet)
        _constrs_contained_within(
            self,
            [
                (
                    None,
                    sum(m.Y[:].get_associated_binary())
                    - (1 - Y_aug[1].get_associated_binary() + 2),
                    0,
                ),
                (
                    1,
                    (1 - m.Y[1].get_associated_binary())
                    + Y_aug[1].get_associated_binary(),
                    None,
                ),
                (
                    None,
                    3
                    - 3 * Y_aug[1].get_associated_binary()
                    - sum(m.Y[:].get_associated_binary()),
                    0,
                ),
            ],
            m.logic_to_linear.transformed_constraints,
        )

        m = ConcreteModel()
        m.s = RangeSet(3)
        m.Y = BooleanVar(m.s)
        m.p = LogicalConstraint(expr=m.Y[1].implies(exactly(2, m.Y[1], m.Y[2], m.Y[3])))
        TransformationFactory('core.logical_to_linear').apply_to(m)
        Y_aug = m.logic_to_linear.augmented_vars
        self.assertEqual(len(Y_aug), 3)
        self.assertEqual(Y_aug[1].domain, BooleanSet)
        _constrs_contained_within(
            self,
            [
                (
                    1,
                    (1 - m.Y[1].get_associated_binary())
                    + Y_aug[1].get_associated_binary(),
                    None,
                ),
                (
                    None,
                    sum(m.Y[:].get_associated_binary())
                    - (1 - Y_aug[1].get_associated_binary() + 2),
                    0,
                ),
                (
                    None,
                    2
                    - 2 * (1 - Y_aug[1].get_associated_binary())
                    - sum(m.Y[:].get_associated_binary()),
                    0,
                ),
                (1, sum(Y_aug[:].get_associated_binary()), None),
                (
                    None,
                    sum(m.Y[:].get_associated_binary())
                    - (1 + 2 * (1 - Y_aug[2].get_associated_binary())),
                    0,
                ),
                (
                    None,
                    3
                    - 3 * (1 - Y_aug[3].get_associated_binary())
                    - sum(m.Y[:].get_associated_binary()),
                    0,
                ),
            ],
            m.logic_to_linear.transformed_constraints,
        )

        # Note: x is now a variable
        m = ConcreteModel()
        m.s = RangeSet(3)
        m.Y = BooleanVar(m.s)
        m.x = Var(bounds=(1, 3))
        m.p = LogicalConstraint(
            expr=m.Y[1].implies(exactly(m.x, m.Y[1], m.Y[2], m.Y[3]))
        )
        TransformationFactory('core.logical_to_linear').apply_to(m)
        Y_aug = m.logic_to_linear.augmented_vars
        self.assertEqual(len(Y_aug), 3)
        self.assertEqual(Y_aug[1].domain, BooleanSet)
        _constrs_contained_within(
            self,
            [
                (
                    1,
                    (1 - m.Y[1].get_associated_binary())
                    + Y_aug[1].get_associated_binary(),
                    None,
                ),
                (
                    None,
                    sum(m.Y[:].get_associated_binary())
                    - (m.x + 2 * (1 - Y_aug[1].get_associated_binary())),
                    0,
                ),
                (
                    None,
                    m.x
                    - 3 * (1 - Y_aug[1].get_associated_binary())
                    - sum(m.Y[:].get_associated_binary()),
                    0,
                ),
                (1, sum(Y_aug[:].get_associated_binary()), None),
                (
                    None,
                    sum(m.Y[:].get_associated_binary())
                    - (m.x - 1 + 3 * (1 - Y_aug[2].get_associated_binary())),
                    0,
                ),
                (
                    None,
                    m.x
                    + 1
                    - 4 * (1 - Y_aug[3].get_associated_binary())
                    - sum(m.Y[:].get_associated_binary()),
                    0,
                ),
            ],
            m.logic_to_linear.transformed_constraints,
        )

    def test_xfrm_atleast_nested(self):
        m = _generate_boolean_model(4)
        m.p = LogicalConstraint(
            expr=atleast(
                1, atleast(2, m.Y[1], m.Y[1].lor(m.Y[2]), m.Y[2]).lor(m.Y[3]), m.Y[4]
            )
        )
        TransformationFactory('core.logical_to_linear').apply_to(m)
        Y_aug = m.logic_to_linear.augmented_vars
        self.assertEqual(len(Y_aug), 3)
        _constrs_contained_within(
            self,
            [
                (
                    1,
                    Y_aug[1].get_associated_binary() + m.Y[4].get_associated_binary(),
                    None,
                ),
                (
                    1,
                    1
                    - Y_aug[2].get_associated_binary()
                    + Y_aug[1].get_associated_binary(),
                    None,
                ),
                (
                    1,
                    1
                    - m.Y[3].get_associated_binary()
                    + Y_aug[1].get_associated_binary(),
                    None,
                ),
                (
                    1,
                    Y_aug[2].get_associated_binary()
                    + m.Y[3].get_associated_binary()
                    + 1
                    - Y_aug[1].get_associated_binary(),
                    None,
                ),
                (
                    1,
                    1
                    - m.Y[1].get_associated_binary()
                    + Y_aug[3].get_associated_binary(),
                    None,
                ),
                (
                    1,
                    1
                    - m.Y[2].get_associated_binary()
                    + Y_aug[3].get_associated_binary(),
                    None,
                ),
                (
                    1,
                    m.Y[1].get_associated_binary()
                    + m.Y[2].get_associated_binary()
                    + 1
                    - Y_aug[3].get_associated_binary(),
                    None,
                ),
                (
                    None,
                    2
                    - 2 * (1 - Y_aug[2].get_associated_binary())
                    - (
                        m.Y[1].get_associated_binary()
                        + Y_aug[3].get_associated_binary()
                        + m.Y[2].get_associated_binary()
                    ),
                    0,
                ),
                (
                    None,
                    m.Y[1].get_associated_binary()
                    + Y_aug[3].get_associated_binary()
                    + m.Y[2].get_associated_binary()
                    - (1 + 2 * Y_aug[2].get_associated_binary()),
                    0,
                ),
            ],
            m.logic_to_linear.transformed_constraints,
        )

    def test_link_with_gdp_indicators(self):
        m = _generate_boolean_model(4)
        m.d1 = Disjunct()
        m.d2 = Disjunct()
        m.x = Var()
        m.dd = Disjunct([1, 2])
        m.d1.c = Constraint(expr=m.x >= 2)
        m.d2.c = Constraint(expr=m.x <= 10)
        m.dd[1].c = Constraint(expr=m.x >= 5)
        m.dd[2].c = Constraint(expr=m.x <= 6)
        m.Y[1].associate_binary_var(m.d1.binary_indicator_var)
        m.Y[2].associate_binary_var(m.d2.binary_indicator_var)
        m.Y[3].associate_binary_var(m.dd[1].binary_indicator_var)
        m.Y[4].associate_binary_var(m.dd[2].binary_indicator_var)
        m.p = LogicalConstraint(expr=m.Y[1].implies(lor(m.Y[3], m.Y[4])))
        m.p2 = LogicalConstraint(expr=atmost(2, *m.Y[:]))
        TransformationFactory('core.logical_to_linear').apply_to(m)
        _constrs_contained_within(
            self,
            [
                (
                    1,
                    m.dd[1].binary_indicator_var
                    + m.dd[2].binary_indicator_var
                    + 1
                    - m.d1.binary_indicator_var,
                    None,
                ),
                (
                    None,
                    m.d1.binary_indicator_var
                    + m.d2.binary_indicator_var
                    + m.dd[1].binary_indicator_var
                    + m.dd[2].binary_indicator_var,
                    2,
                ),
            ],
            m.logic_to_linear.transformed_constraints,
        )

    def test_gdp_nesting(self):
        m = _generate_boolean_model(2)
        m.disj = Disjunction(
            expr=[[m.Y[1].implies(m.Y[2])], [m.Y[2].equivalent_to(False)]]
        )
        TransformationFactory('core.logical_to_linear').apply_to(
            m, targets=[m.disj.disjuncts[0], m.disj.disjuncts[1]]
        )
        _constrs_contained_within(
            self,
            [
                (
                    1,
                    1 - m.Y[1].get_associated_binary() + m.Y[2].get_associated_binary(),
                    None,
                )
            ],
            m.disj_disjuncts[0].logic_to_linear.transformed_constraints,
        )
        _constrs_contained_within(
            self,
            [(1, 1 - m.Y[2].get_associated_binary(), 1)],
            m.disj_disjuncts[1].logic_to_linear.transformed_constraints,
        )

    def test_transformed_components_on_parent_block(self):
        m = ConcreteModel()
        m.b = Block()
        m.b.s = RangeSet(3)
        m.b.Y = BooleanVar(m.b.s)
        m.b.p = LogicalConstraint(expr=m.b.Y[1].implies(lor(m.b.Y[2], m.b.Y[3])))
        TransformationFactory('core.logical_to_linear').apply_to(m)

        boolean_var = m.b.component("Y_asbinary")
        self.assertIsInstance(boolean_var, Var)
        notAVar = m.component("Y_asbinary")
        self.assertIsNone(notAVar)

        transBlock = m.b.component("logic_to_linear")
        self.assertIsInstance(transBlock, Block)
        notAThing = m.component("logic_to_linear")
        self.assertIsNone(notAThing)

        # check the constraints on the transBlock
        _constrs_contained_within(
            self,
            [
                (
                    1,
                    m.b.Y[2].get_associated_binary()
                    + m.b.Y[3].get_associated_binary()
                    + (1 - m.b.Y[1].get_associated_binary()),
                    None,
                )
            ],
            m.b.logic_to_linear.transformed_constraints,
        )

    def make_nested_block_model(self):
        """For the next two tests: Has BooleanVar on model, but
        LogicalConstraints on a Block and a Block nested on that Block."""
        m = ConcreteModel()
        m.b = Block()
        m.Y = BooleanVar([1, 2])
        m.b.logical = LogicalConstraint(expr=~m.Y[1])
        m.b.b = Block()
        m.b.b.logical = LogicalConstraint(expr=m.Y[1].xor(m.Y[2]))
        return m

    def test_transform_block(self):
        m = self.make_nested_block_model()
        TransformationFactory('core.logical_to_linear').apply_to(m.b)

        _constrs_contained_within(
            self,
            [(1, 1 - m.Y[1].get_associated_binary(), 1)],
            m.b.logic_to_linear.transformed_constraints,
        )
        # ESJ: This is kinda whacky looking... Why not Y[1] + Y[2] == 1? (It's
        # special case of an exactly(1, ...) constraint.
        _constrs_contained_within(
            self,
            [
                (
                    1,
                    m.Y[1].get_associated_binary() + m.Y[2].get_associated_binary(),
                    None,
                ),
                (
                    1,
                    1
                    - m.Y[1].get_associated_binary()
                    + 1
                    - m.Y[2].get_associated_binary(),
                    None,
                ),
            ],
            m.b.b.logic_to_linear.transformed_constraints,
        )
        self.assertEqual(len(m.b.logic_to_linear.transformed_constraints), 1)
        self.assertEqual(len(m.b.b.logic_to_linear.transformed_constraints), 2)

    def test_transform_targets_on_block(self):
        m = self.make_nested_block_model()
        TransformationFactory('core.logical_to_linear').apply_to(m.b, targets=m.b.b)
        # didn't transform anything on m.b
        self.assertIsNone(m.b.component("logic_to_linear"))
        # got what we expected on m.b.b
        _constrs_contained_within(
            self,
            [
                (
                    1,
                    m.Y[1].get_associated_binary() + m.Y[2].get_associated_binary(),
                    None,
                ),
                (
                    1,
                    1
                    - m.Y[1].get_associated_binary()
                    + 1
                    - m.Y[2].get_associated_binary(),
                    None,
                ),
            ],
            m.b.b.logic_to_linear.transformed_constraints,
        )
        self.assertEqual(len(m.b.b.logic_to_linear.transformed_constraints), 2)

    def test_logical_constraint_target(self):
        m = _generate_boolean_model(3)
        TransformationFactory('core.logical_to_linear').apply_to(
            m, targets=m.constraint
        )
        _constrs_contained_within(
            self,
            [
                (
                    2,
                    m.Y[1].get_associated_binary()
                    + m.Y[2].get_associated_binary()
                    + m.Y[3].get_associated_binary(),
                    2,
                )
            ],
            m.logic_to_linear.transformed_constraints,
        )

    def make_indexed_logical_constraint_model(self):
        m = _generate_boolean_model(3)
        m.cons = LogicalConstraint([1, 2])
        m.cons[1] = exactly(2, m.Y)
        m.cons[2] = m.Y[1].implies(lor(m.Y[2], m.Y[3]))
        return m

    def test_indexed_logical_constraint_target(self):
        m = self.make_indexed_logical_constraint_model()
        TransformationFactory('core.logical_to_linear').apply_to(m, targets=m.cons)
        _constrs_contained_within(
            self,
            [
                (
                    2,
                    m.Y[1].get_associated_binary()
                    + m.Y[2].get_associated_binary()
                    + m.Y[3].get_associated_binary(),
                    2,
                )
            ],
            m.logic_to_linear.transformed_constraints,
        )
        _constrs_contained_within(
            self,
            [
                (
                    1,
                    m.Y[2].get_associated_binary()
                    + m.Y[3].get_associated_binary()
                    + (1 - m.Y[1].get_associated_binary()),
                    None,
                )
            ],
            m.logic_to_linear.transformed_constraints,
        )

        # and verify only the targets were transformed
        self.assertEqual(len(m.logic_to_linear.transformed_constraints), 2)
        self.assertTrue(m.constraint.active)

    def test_logical_constraintData_target(self):
        m = self.make_indexed_logical_constraint_model()
        TransformationFactory('core.logical_to_linear').apply_to(m, targets=m.cons[2])
        _constrs_contained_within(
            self,
            [
                (
                    1,
                    m.Y[2].get_associated_binary()
                    + m.Y[3].get_associated_binary()
                    + (1 - m.Y[1].get_associated_binary()),
                    None,
                )
            ],
            m.logic_to_linear.transformed_constraints,
        )
        # only transformed the second one.
        self.assertEqual(len(m.logic_to_linear.transformed_constraints), 1)

    def test_blockData_target(self):
        m = ConcreteModel()
        m.b = Block([1, 2])
        m.b[1].transfer_attributes_from(self.make_indexed_logical_constraint_model())
        TransformationFactory('core.logical_to_linear').apply_to(m, targets=m.b[1])
        _constrs_contained_within(
            self,
            [
                (
                    2,
                    m.b[1].Y[1].get_associated_binary()
                    + m.b[1].Y[2].get_associated_binary()
                    + m.b[1].Y[3].get_associated_binary(),
                    2,
                )
            ],
            m.b[1].logic_to_linear.transformed_constraints,
        )
        _constrs_contained_within(
            self,
            [
                (
                    1,
                    m.b[1].Y[2].get_associated_binary()
                    + m.b[1].Y[3].get_associated_binary()
                    + (1 - m.b[1].Y[1].get_associated_binary()),
                    None,
                )
            ],
            m.b[1].logic_to_linear.transformed_constraints,
        )

    def test_disjunctData_target(self):
        m = ConcreteModel()
        m.d = Disjunct([1, 2])
        m.d[1].transfer_attributes_from(self.make_indexed_logical_constraint_model())
        TransformationFactory('core.logical_to_linear').apply_to(m, targets=m.d[1])
        _constrs_contained_within(
            self,
            [
                (
                    2,
                    m.d[1].Y[1].get_associated_binary()
                    + m.d[1].Y[2].get_associated_binary()
                    + m.d[1].Y[3].get_associated_binary(),
                    2,
                )
            ],
            m.d[1].logic_to_linear.transformed_constraints,
        )
        _constrs_contained_within(
            self,
            [
                (
                    1,
                    m.d[1].Y[2].get_associated_binary()
                    + m.d[1].Y[3].get_associated_binary()
                    + (1 - m.d[1].Y[1].get_associated_binary()),
                    None,
                )
            ],
            m.d[1].logic_to_linear.transformed_constraints,
        )

    def test_target_with_unrecognized_type(self):
        m = _generate_boolean_model(2)
        with self.assertRaisesRegex(
            ValueError,
            r"invalid value for configuration "
            r"'targets':\n\tFailed casting 1\n\tto "
            r"target_list\n\tError: "
            r"Expected Component or list of Components."
            r"\n\tReceived <class 'int'>",
        ):
            TransformationFactory('core.logical_to_linear').apply_to(m, targets=1)

    def test_mixed_logical_relational_expressions(self):
        m = ConcreteModel()
        m.x = Var()
        m.y = BooleanVar([1, 2])
        m.c = LogicalConstraint(expr=(land(m.y[1], m.y[2]).implies(m.x >= 0)))
        with self.assertRaisesRegex(
            MouseTrap,
            "core.logical_to_linear does not support transforming "
            "LogicalConstraints with embedded relational expressions. "
            "Found '0.0 <= x'.",
            normalize_whitespace=True,
        ):
            TransformationFactory('core.logical_to_linear').apply_to(m)

    def test_external_function(self):
        def _fcn(*args):
            raise RuntimeError('unreachable')

        m = ConcreteModel()
        m.x = Var()
        m.f = ExternalFunction(_fcn)
        m.y = BooleanVar()
        m.c = LogicalConstraint(expr=(m.y.implies(m.f(m.x))))
        with self.assertRaisesRegex(
            TypeError,
            "Expressions containing external functions are not convertible "
            r"to sympy expressions \(found 'f\(x1",
        ):
            TransformationFactory('core.logical_to_linear').apply_to(m)


@unittest.skipUnless(sympy_available, "Sympy not available")
class TestLogicalToLinearBackmap(unittest.TestCase):
    def test_backmap_deprecated(self):
        m = ConcreteModel()
        m.s = RangeSet(3)
        m.Y = BooleanVar(m.s)
        TransformationFactory('core.logical_to_linear').apply_to(m)
        output = StringIO()
        with LoggingIntercept(output, 'pyomo.core.base', logging.WARNING):
            y1 = m.Y[1].get_associated_binary()
        self.assertIn(
            "DEPRECATED: Relying on core.logical_to_linear to "
            "transform BooleanVars that do not appear in "
            "LogicalConstraints is deprecated. Please "
            "associate your own binaries if you have BooleanVars "
            "not used in logical expressions.",
            output.getvalue().replace('\n', ' '),
        )
        output = StringIO()
        with LoggingIntercept(output, 'pyomo.core.base', logging.WARNING):
            y2 = m.Y[2].get_associated_binary()
        self.assertIn(
            "DEPRECATED: Relying on core.logical_to_linear to "
            "transform BooleanVars that do not appear in "
            "LogicalConstraints is deprecated. Please "
            "associate your own binaries if you have BooleanVars "
            "not used in logical expressions.",
            output.getvalue().replace('\n', ' '),
        )
        y1.value = 1
        y2.value = 0
        update_boolean_vars_from_binary(m)
        self.assertTrue(m.Y[1].value)
        self.assertFalse(m.Y[2].value)
        self.assertIsNone(m.Y[3].value)

    def test_can_associate_unused_boolean_after_transformation(self):
        m = ConcreteModel()
        m.Y = BooleanVar()
        TransformationFactory('core.logical_to_linear').apply_to(m)
        m.y = Var(domain=Binary)
        output = StringIO()
        with LoggingIntercept(output, 'pyomo.core.base', logging.WARNING):
            m.Y.associate_binary_var(m.y)
            y = m.Y.get_associated_binary()
        self.assertIs(y, m.y)
        # we didn't whine about this
        self.assertEqual(output.getvalue(), '')

    def test_cannot_reassociate_boolean_error(self):
        m = _generate_boolean_model(2)
        TransformationFactory('core.logical_to_linear').apply_to(m)
        # both of the variable have been associated with binaries, we're not
        # allowed to change now.
        m.y = Var(domain=Binary)
        with self.assertRaisesRegex(
            RuntimeError,
            r"Reassociating BooleanVar 'Y\[1\]' "
            r"\(currently associated with 'Y_asbinary\[1\]'\)"
            r" with 'y' is not allowed",
        ):
            m.Y[1].associate_binary_var(m.y)

    def test_backmap(self):
        m = _generate_boolean_model(3)
        TransformationFactory('core.logical_to_linear').apply_to(m)
        m.Y_asbinary[1].value = 1
        m.Y_asbinary[2].value = 0
        update_boolean_vars_from_binary(m)
        self.assertTrue(m.Y[1].value)
        self.assertFalse(m.Y[2].value)
        self.assertIsNone(m.Y[3].value)

    def test_backmap_hierarchical_model(self):
        m = _generate_boolean_model(3)
        m.b = Block()
        m.b.Y = BooleanVar()
        m.b.lc = LogicalConstraint(expr=m.Y[1].lor(m.b.Y))
        TransformationFactory('core.logical_to_linear').apply_to(m)
        m.Y_asbinary[1].value = 1
        m.Y_asbinary[2].value = 0
        m.b.Y.get_associated_binary().value = 1
        update_boolean_vars_from_binary(m)
        self.assertTrue(m.Y[1].value)
        self.assertFalse(m.Y[2].value)
        self.assertIsNone(m.Y[3].value)
        self.assertTrue(m.b.Y.value)

    def test_backmap_noninteger(self):
        m = _generate_boolean_model(2)
        TransformationFactory('core.logical_to_linear').apply_to(m)
        m.Y_asbinary[1].value = 0.9
        update_boolean_vars_from_binary(m, integer_tolerance=0.1)
        self.assertTrue(m.Y[1].value)
        # Now try it without the tolerance set
        with self.assertRaisesRegex(
            ValueError, r"Binary variable has non-\{0,1\} value"
        ):
            update_boolean_vars_from_binary(m)


if __name__ == "__main__":
    unittest.main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

#
# Unit Tests for ContinuousSet() Objects
#

import os
from os.path import abspath, dirname

import pyomo.common.unittest as unittest

from pyomo.environ import ConcreteModel, AbstractModel, Set
from pyomo.dae import ContinuousSet
from pyomo.common.log import LoggingIntercept
from io import StringIO

currdir = dirname(abspath(__file__)) + os.sep


class TestContinuousSet(unittest.TestCase):
    # test __init__
    def test_init(self):
        model = ConcreteModel()
        model.t = ContinuousSet(bounds=(0, 1))

        model = ConcreteModel()
        model.t = ContinuousSet(initialize=[1, 2, 3])

        model = ConcreteModel()
        model.t = ContinuousSet(bounds=(0, 5), initialize=[1, 3, 5])

        # Expected ValueError because a ContinuousSet component
        # must contain at least two values upon construction
        with self.assertRaises(ValueError):
            model.t = ContinuousSet()

    # test bad keyword arguments
    def test_bad_kwds(self):
        model = ConcreteModel()
        with self.assertRaises(TypeError):
            model.t = ContinuousSet(bounds=(0, 1), filter=True)

        with self.assertRaises(TypeError):
            model.t = ContinuousSet(bounds=(0, 1), dimen=2)

        with self.assertRaises(TypeError):
            model.t = ContinuousSet(bounds=(0, 1), virtual=True)

        with self.assertRaises(TypeError):
            model.t = ContinuousSet(bounds=(0, 1), validate=True)

    # test valid declarations
    def test_valid_declaration(self):
        model = ConcreteModel()
        model.t = ContinuousSet(bounds=(0, 1))
        self.assertEqual(len(model.t), 2)
        self.assertIn(0, model.t)
        self.assertIn(1, model.t)

        model = ConcreteModel()
        model.t = ContinuousSet(initialize=[1, 2, 3])
        self.assertEqual(len(model.t), 3)
        self.assertEqual(model.t.first(), 1)
        self.assertEqual(model.t.last(), 3)

        model = ConcreteModel()
        model.t = ContinuousSet(bounds=(1, 3), initialize=[1, 2, 3])
        self.assertEqual(len(model.t), 3)
        self.assertEqual(model.t.first(), 1)
        self.assertEqual(model.t.last(), 3)

        model = ConcreteModel()
        model.t = ContinuousSet(bounds=(0, 4), initialize=[1, 2, 3])
        self.assertEqual(len(model.t), 5)
        self.assertEqual(model.t.first(), 0)
        self.assertEqual(model.t.last(), 4)

        model = ConcreteModel()
        with self.assertRaisesRegex(ValueError, r"value is not in the domain \[0..4\]"):
            model.t = ContinuousSet(bounds=(0, 4), initialize=[1, 2, 3, 5])
        # self.assertEqual(len(model.t), 5)
        # self.assertEqual(model.t.first(), 0)
        # self.assertEqual(model.t.last(), 5)
        # self.assertNotIn(4, model.t)
        # del model.t

        model = ConcreteModel()
        with self.assertRaisesRegex(ValueError, r"value is not in the domain \[2..6\]"):
            model.t = ContinuousSet(bounds=(2, 6), initialize=[1, 2, 3, 5])
        # self.assertEqual(len(model.t), 5)
        # self.assertEqual(model.t.first(), 1)
        # self.assertEqual(model.t.last(), 6)
        # del model.t

        model = ConcreteModel()
        with self.assertRaisesRegex(ValueError, r"value is not in the domain \[2..4\]"):
            model.t = ContinuousSet(bounds=(2, 4), initialize=[1, 3, 5])
        # self.assertEqual(len(model.t), 3)
        # self.assertNotIn(2, model.t)
        # self.assertNotIn(4, model.t)

    # test invalid declarations
    def test_invalid_declaration(self):
        model = ConcreteModel()
        model.s = Set(initialize=[1, 2, 3])
        with self.assertRaises(TypeError):
            model.t = ContinuousSet(model.s, bounds=(0, 1))

        model = ConcreteModel()
        with self.assertRaises(ValueError):
            model.t = ContinuousSet(bounds=(0, 0))

        model = ConcreteModel()
        with self.assertRaises(ValueError):
            model.t = ContinuousSet(initialize=[1])

        model = ConcreteModel()
        with self.assertRaises(ValueError):
            model.t = ContinuousSet(bounds=(None, 1))

        model = ConcreteModel()
        with self.assertRaises(ValueError):
            model.t = ContinuousSet(bounds=(0, None))

        model = ConcreteModel()
        with self.assertRaises(ValueError):
            model.t = ContinuousSet(initialize=[(1, 2), (3, 4)])

        model = ConcreteModel()
        with self.assertRaises(ValueError):
            model.t = ContinuousSet(initialize=['foo', 'bar'])

    # test the get_changed method
    def test_get_changed(self):
        model = ConcreteModel()
        model.t = ContinuousSet(initialize=[1, 2, 3])
        self.assertFalse(model.t.get_changed())
        self.assertEqual(model.t._changed, model.t.get_changed())

    # test the set_changed method
    def test_set_changed(self):
        model = ConcreteModel()
        model.t = ContinuousSet(initialize=[1, 2, 3])
        self.assertFalse(model.t._changed)
        model.t.set_changed(True)
        self.assertTrue(model.t._changed)
        model.t.set_changed(False)
        self.assertFalse(model.t._changed)

        with self.assertRaises(ValueError):
            model.t.set_changed(3)

    # test get_upper_element_boundary
    def test_get_upper_element_boundary(self):
        m = ConcreteModel()
        m.t = ContinuousSet(initialize=[1, 2, 3])
        self.assertEqual(m.t.get_upper_element_boundary(1.5), 2)
        self.assertEqual(m.t.get_upper_element_boundary(2.5), 3)
        self.assertEqual(m.t.get_upper_element_boundary(2), 2)

        log_out = StringIO()
        with LoggingIntercept(log_out, 'pyomo.dae'):
            temp = m.t.get_upper_element_boundary(3.5)
        self.assertIn('Returning the upper bound', log_out.getvalue())

    # test get_lower_element_boundary
    def test_get_lower_element_boundary(self):
        m = ConcreteModel()
        m.t = ContinuousSet(initialize=[1, 2, 3])
        self.assertEqual(m.t.get_lower_element_boundary(1.5), 1)
        self.assertEqual(m.t.get_lower_element_boundary(2.5), 2)
        self.assertEqual(m.t.get_lower_element_boundary(2), 2)

        log_out = StringIO()
        with LoggingIntercept(log_out, 'pyomo.dae'):
            temp = m.t.get_lower_element_boundary(0.5)
        self.assertIn('Returning the lower bound', log_out.getvalue())

    def test_duplicate_construct(self):
        m = ConcreteModel()
        m.t = ContinuousSet(initialize=[1, 2, 3])
        self.assertEqual(m.t, [1, 2, 3])
        self.assertEqual(m.t._fe, [1, 2, 3])
        m.t.add(1.5)
        m.t.add(2.5)
        self.assertEqual(m.t, [1, 1.5, 2, 2.5, 3])
        self.assertEqual(m.t._fe, [1, 2, 3])
        m.t.construct()
        self.assertEqual(m.t, [1, 1.5, 2, 2.5, 3])
        self.assertEqual(m.t._fe, [1, 2, 3])

    def test_find_nearest_index(self):
        m = ConcreteModel()
        m.time = ContinuousSet(bounds=(0, 5))

        # Undiscretized
        i = m.time.find_nearest_index(1)
        self.assertEqual(i, 1)
        i = m.time.find_nearest_index(1, tolerance=0.5)
        self.assertEqual(i, None)

        i = m.time.find_nearest_index(-0.01, tolerance=0.1)
        self.assertEqual(i, 1)
        i = m.time.find_nearest_index(-0.01, tolerance=0.001)
        self.assertEqual(i, None)

        i = m.time.find_nearest_index(6, tolerance=2)
        self.assertEqual(i, 2)
        i = m.time.find_nearest_index(6, tolerance=1)
        self.assertEqual(i, 2)

        # This test relies on the behavior for tiebreaks
        i = m.time.find_nearest_index(2.5)
        self.assertEqual(i, 1)

        m.del_component(m.time)
        init_list = []
        for i in range(5):
            i0 = float(i)
            i1 = round((i + 0.15) * 1e4) / 1e4
            i2 = round((i + 0.64) * 1e4) / 1e4
            # Round to get rid of numerical error due to float addition
            init_list.extend([i, i1, i2])
        init_list.append(5.0)
        m.time = ContinuousSet(initialize=init_list)

        # "Discretized"
        i = m.time.find_nearest_index(1.01, tolerance=0.1)
        self.assertEqual(i, 4)
        i = m.time.find_nearest_index(1.01, tolerance=0.001)
        self.assertEqual(i, None)

        i = m.time.find_nearest_index(3.5)
        self.assertEqual(i, 12)
        i = m.time.find_nearest_index(3.5, tolerance=0.1)
        self.assertEqual(i, None)

        i = m.time.find_nearest_index(-1)
        self.assertEqual(i, 1)
        i = m.time.find_nearest_index(-1, tolerance=1)
        self.assertEqual(i, 1)

        i = m.time.find_nearest_index(5.5)
        self.assertEqual(i, 16)
        i = m.time.find_nearest_index(5.5, tolerance=0.49)
        self.assertEqual(i, None)

        i = m.time.find_nearest_index(2.64, tolerance=1e-8)
        self.assertEqual(i, 9)
        i = m.time.find_nearest_index(2.64, tolerance=0)
        self.assertEqual(i, 9)

        i = m.time.find_nearest_index(5, tolerance=0)
        self.assertEqual(i, 16)

        i = m.time.find_nearest_index(0, tolerance=0)
        self.assertEqual(i, 1)

        # This test fails. I get:
        # delta_left == 2.075-2.0 == 0.07500000000000018
        # delta_right == 2.15-2.075 == 0.07499999999999973
        # Index 8 is returned (time[8] == 2.15), even though
        # "tie"-break logic suggests index 7 should be.
        # i = m.time.find_nearest_index(2.075)
        # self.assertEqual(i, 7)


class TestIO(unittest.TestCase):
    def setUp(self):
        #
        # Create Model
        #
        self.model = AbstractModel()
        self.instance = None

    def tearDown(self):
        if os.path.exists("diffset.dat"):
            os.remove("diffset.dat")
        self.model = None
        self.instance = None

    def test_io1(self):
        OUTPUT = open("diffset.dat", "w")
        OUTPUT.write("data;\n")
        OUTPUT.write("set A := 1 3 5 7;\n")
        OUTPUT.write("end;\n")
        OUTPUT.close()
        self.model.A = ContinuousSet()
        self.instance = self.model.create_instance("diffset.dat")
        self.assertEqual(len(self.instance.A), 4)

    def test_io2(self):
        OUTPUT = open("diffset.dat", "w")
        OUTPUT.write("data;\n")
        OUTPUT.write("set A := 1 3 5;\n")
        OUTPUT.write("end;\n")
        OUTPUT.close()
        self.model.A = ContinuousSet(bounds=(0, 4))
        with self.assertRaisesRegex(
            ValueError, r"The value is not in the domain \[0..4\]"
        ):
            self.instance = self.model.create_instance("diffset.dat")
        # self.assertEqual(len(self.instance.A), 4)

    def test_io3(self):
        OUTPUT = open("diffset.dat", "w")
        OUTPUT.write("data;\n")
        OUTPUT.write("set A := 1 3 5;\n")
        OUTPUT.write("end;\n")
        OUTPUT.close()
        self.model.A = ContinuousSet(bounds=(2, 6))
        with self.assertRaisesRegex(
            ValueError, r"The value is not in the domain \[2..6\]"
        ):
            self.instance = self.model.create_instance("diffset.dat")
        # self.assertEqual(len(self.instance.A), 4)

    def test_io4(self):
        OUTPUT = open("diffset.dat", "w")
        OUTPUT.write("data;\n")
        OUTPUT.write("set A := 1 3 5;\n")
        OUTPUT.write("end;\n")
        OUTPUT.close()
        self.model.A = ContinuousSet(bounds=(2, 4))
        with self.assertRaisesRegex(
            ValueError, r"The value is not in the domain \[2..4\]"
        ):
            self.instance = self.model.create_instance("diffset.dat")
        # self.assertEqual(len(self.instance.A), 3)

    def test_io5(self):
        OUTPUT = open("diffset.dat", "w")
        OUTPUT.write("data;\n")
        OUTPUT.write("set A := 1 3 5;\n")
        OUTPUT.write("end;\n")
        OUTPUT.close()
        self.model.A = ContinuousSet(bounds=(0, 6))
        self.instance = self.model.create_instance("diffset.dat")
        self.assertEqual(len(self.instance.A), 5)

    def test_io6(self):
        OUTPUT = open("diffset.dat", "w")
        OUTPUT.write("data;\n")
        OUTPUT.write("set B := 1;\n")
        OUTPUT.write("end;\n")
        OUTPUT.close()

        # Expected ValueError because data set has only one value and no
        # bounds are specified
        self.model.B = ContinuousSet()
        with self.assertRaises(ValueError):
            self.instance = self.model.create_instance("diffset.dat")

    def test_io7(self):
        OUTPUT = open("diffset.dat", "w")
        OUTPUT.write("data;\n")
        OUTPUT.write("set B := 1;\n")
        OUTPUT.write("end;\n")
        OUTPUT.close()
        self.model.B = ContinuousSet(bounds=(0, 1))
        self.instance = self.model.create_instance("diffset.dat")
        self.assertEqual(len(self.instance.B), 2)


if __name__ == "__main__":
    unittest.main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________
#
# Author:  Gabe Hackebeil
# Purpose: For regression testing to ensure that the Pyomo
#          NL writer properly reports the values corresponding
#          to the nl file header line with the label
#          '# nonlinear vars in constraints, objectives, both'
#
# Test if variables in deactivated blocks are found
#

from pyomo.environ import ConcreteModel, Var, Block, Objective, Constraint

model = ConcreteModel()

model.x = Var(initialize=1.0)
model.b = Block()
model.b.y = Var(initialize=1.0)

model.OBJ = Objective(expr=model.x**2)

model.CON1 = Constraint(expr=model.b.y**2 == 4)

model.b.deactivate()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from .base_linear_solver_interface import IPLinearSolverInterface
from pyomo.contrib.pynumero.linalg.base import LinearSolverStatus, LinearSolverResults
from pyomo.common.dependencies import attempt_import
from collections import OrderedDict
from typing import Union, Optional, Tuple
from pyomo.contrib.pynumero.sparse import BlockVector
import numpy as np

mumps, mumps_available = attempt_import(
    name='pyomo.contrib.pynumero.linalg.mumps_interface',
    error_message='pymumps is required to use the MumpsInterface',
)

from pyomo.contrib.pynumero.linalg.mumps_interface import (
    MumpsCentralizedAssembledLinearSolver,
)


class MumpsInterface(MumpsCentralizedAssembledLinearSolver, IPLinearSolverInterface):
    @classmethod
    def getLoggerName(cls):
        return 'mumps'

    def __init__(self, par=1, comm=None, cntl_options=None, icntl_options=None):
        if icntl_options is None:
            icntl_options = dict()

        # These options are set in order to get the correct inertia.
        if 13 not in icntl_options:
            icntl_options[13] = 1
        if 24 not in icntl_options:
            icntl_options[24] = 0

        super(MumpsInterface, self).__init__(
            sym=2,
            par=par,
            comm=comm,
            cntl_options=cntl_options,
            icntl_options=icntl_options,
        )

        self.error_level = self.get_icntl(11)
        self.log_error = bool(self.error_level)
        self.logger = self.getLogger()
        self.log_header(include_error=self.log_error)

    def do_back_solve(
        self, rhs: Union[np.ndarray, BlockVector], raise_on_error: bool = True
    ) -> Tuple[Optional[Union[np.ndarray, BlockVector]], LinearSolverResults]:
        res, status = super(MumpsInterface, self).do_back_solve(
            rhs, raise_on_error=raise_on_error
        )
        self.log_info()
        return res, status

    def get_inertia(self):
        num_negative_eigenvalues = self.get_infog(12)
        num_zero_eigenvalues = self.get_infog(28)
        num_positive_eigenvalues = (
            self._dim - num_negative_eigenvalues - num_zero_eigenvalues
        )
        return num_positive_eigenvalues, num_negative_eigenvalues, num_zero_eigenvalues

    def get_error_info(self):
        # Access error level contained in ICNTL(11) (Fortran indexing).
        # Assuming this value has not changed since the solve was performed.
        error_level = self.get_icntl(11)
        info = OrderedDict()
        if error_level == 0:
            return info
        elif error_level == 1:
            info['||A||'] = self.get_rinfog(4)
            info['||x||'] = self.get_rinfog(5)
            info['Max resid'] = self.get_rinfog(6)
            info['Max error'] = self.get_rinfog(9)
            return info
        elif error_level == 2:
            info['||A||'] = self.get_rinfog(4)
            info['||x||'] = self.get_rinfog(5)
            info['Max resid'] = self.get_rinfog(6)
            return info

    def log_header(self, include_error=True, extra_fields=None):
        if extra_fields is None:
            extra_fields = list()
        header_fields = []
        header_fields.append('Status')
        header_fields.append('n_null')
        header_fields.append('n_neg')

        if include_error:
            header_fields.extend(self.get_error_info().keys())

        header_fields.extend(extra_fields)

        # Allocate 10 spaces for integer values
        header_string = '{0:<10}'
        header_string += '{1:<10}'
        header_string += '{2:<10}'

        # Allocate 15 spaces for the rest, which I assume are floats
        for i in range(4, len(header_fields)):
            header_string += '{' + str(i) + ':<15}'

        self.logger.info(header_string.format(*header_fields))

    def log_info(self):
        # Which fields to log should be specified at the instance level
        # Any logging that should be done on an iteration-specific case
        # should be handled by the IP solver
        fields = []
        fields.append(self.get_infog(1))  # Status, 0 for success
        fields.append(self.get_infog(28))  # Number of null pivots
        fields.append(self.get_infog(12))  # Number of negative pivots

        include_error = self.log_error
        if include_error:
            fields.extend(self.get_error_info().values())

        extra_fields = []
        fields.extend(extra_fields)

        # Allocate 10 spaces for integer values
        log_string = '{0:<10}'
        log_string += '{1:<10}'
        log_string += '{2:<10}'

        # Allocate 15 spaces for the rest, which I assume are floats
        for i in range(4, len(fields)):
            log_string += '{' + str(i) + ':<15.3e}'

        self.logger.info(log_string.format(*fields))



# -*- coding: utf-8 -*-

# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from pyomo.contrib.mindtpy.util import calc_jacobians
from pyomo.core import ConstraintList
from pyomo.opt import SolverFactory
from pyomo.contrib.mindtpy.config_options import _get_MindtPy_OA_config
from pyomo.contrib.mindtpy.algorithm_base_class import _MindtPyAlgorithm
from pyomo.contrib.mindtpy.cut_generation import add_oa_cuts, add_oa_cuts_for_grey_box


@SolverFactory.register(
    'mindtpy.oa', doc='MindtPy: Mixed-Integer Nonlinear Decomposition Toolbox in Pyomo'
)
class MindtPy_OA_Solver(_MindtPyAlgorithm):
    """
    Decomposition solver for Mixed-Integer Nonlinear Programming (MINLP) problems.

    The MindtPy (Mixed-Integer Nonlinear Decomposition Toolbox in Pyomo) solver
    applies a variety of decomposition-based approaches to solve Mixed-Integer
    Nonlinear Programming (MINLP) problems.
    This class includes:

    - Outer approximation (OA)
    - Regularized outer approximation (ROA)
    - LP/NLP based branch-and-bound (LP/NLP)
    - Regularized LP/NLP based branch-and-bound (RLP/NLP)
    """

    CONFIG = _get_MindtPy_OA_config()

    def check_config(self):
        config = self.config
        if config.add_regularization is not None:
            if config.add_regularization in {
                'grad_lag',
                'hess_lag',
                'hess_only_lag',
                'sqp_lag',
            }:
                config.calculate_dual_at_solution = True
            if config.regularization_mip_threads == 0 and config.threads > 0:
                config.regularization_mip_threads = config.threads
                config.logger.info('Set regularization_mip_threads equal to threads')
            if config.single_tree:
                config.add_cuts_at_incumbent = True
            if config.mip_regularization_solver is None:
                config.mip_regularization_solver = config.mip_solver
        if config.single_tree:
            config.logger.info('Single-tree implementation is activated.')
            config.iteration_limit = 1
            config.add_slack = False
            if config.mip_solver not in {'cplex_persistent', 'gurobi_persistent'}:
                raise ValueError(
                    "Only cplex_persistent and gurobi_persistent are supported for LP/NLP based Branch and Bound method."
                    "Please refer to https://pyomo.readthedocs.io/en/stable/explanation/solvers/mindtpy.html#lp-nlp-based-branch-and-bound."
                )
            if config.threads > 1:
                config.threads = 1
                config.logger.info(
                    'The threads parameter is corrected to 1 since lazy constraint callback conflicts with multi-threads mode.'
                )
        if config.heuristic_nonconvex:
            config.equality_relaxation = True
            config.add_slack = True
        if config.equality_relaxation:
            config.calculate_dual_at_solution = True
        if config.init_strategy == 'FP' or config.add_regularization is not None:
            config.move_objective = True
        if config.add_regularization is not None:
            if config.add_regularization in {
                'level_L1',
                'level_L_infinity',
                'grad_lag',
            }:
                self.regularization_mip_type = 'MILP'
            elif config.add_regularization in {
                'level_L2',
                'hess_lag',
                'hess_only_lag',
                'sqp_lag',
            }:
                self.regularization_mip_type = 'MIQP'
        _MindtPyAlgorithm.check_config(self)

    def initialize_mip_problem(self):
        """Deactivate the nonlinear constraints to create the MIP problem."""
        super().initialize_mip_problem()
        self.jacobians = calc_jacobians(
            self.mip.MindtPy_utils.nonlinear_constraint_list,
            self.config.differentiate_mode,
        )  # preload jacobians
        self.mip.MindtPy_utils.cuts.oa_cuts = ConstraintList(
            doc='Outer approximation cuts'
        )

    def add_cuts(
        self,
        dual_values,
        linearize_active=True,
        linearize_violated=True,
        cb_opt=None,
        nlp=None,
    ):
        add_oa_cuts(
            self.mip,
            dual_values,
            self.jacobians,
            self.objective_sense,
            self.mip_constraint_polynomial_degree,
            self.mip_iter,
            self.config,
            self.timing,
            cb_opt,
            linearize_active,
            linearize_violated,
        )
        if len(self.mip.MindtPy_utils.grey_box_list) > 0:
            add_oa_cuts_for_grey_box(
                self.mip, nlp, self.config, self.objective_sense, self.mip_iter, cb_opt
            )

    def deactivate_no_good_cuts_when_fixing_bound(self, no_good_cuts):
        # Only deactivate the last OA cuts may not be correct.
        # Since integer solution may also be cut off by OA cuts due to calculation approximation.
        if self.config.add_no_good_cuts:
            no_good_cuts[len(no_good_cuts)].deactivate()
        if self.config.use_tabu_list:
            self.integer_list = self.integer_list[:-1]

    def objective_reformulation(self):
        # In the process_objective function, as long as the objective function is nonlinear, it will be reformulated and the variable/constraint/objective lists will be updated.
        # For OA/GOA/LP-NLP algorithm, if the objective function is linear, it will not be reformulated as epigraph constraint.
        # If the objective function is linear, it will be reformulated as epigraph constraint only if the Feasibility Pump or ROA/RLP-NLP algorithm is activated. (move_objective = True)
        # In some cases, the variable/constraint/objective lists will not be updated even if the objective is epigraph-reformulated.
        # In Feasibility Pump, since the distance calculation only includes discrete variables and the epigraph slack variables are continuous variables, the Feasibility Pump algorithm will not affected even if the variable list are updated.
        # In ROA and RLP/NLP, since the distance calculation does not include these epigraph slack variables, they should not be added to the variable list. (update_var_con_list = False)
        # In the process_objective function, once the objective function has been reformulated as epigraph constraint, the variable/constraint/objective lists will not be updated only if the MINLP has a linear objective function and regularization is activated at the same time.
        # This is because the epigraph constraint is very "flat" for branching rules. The original objective function will be used for the main problem and epigraph reformulation will be used for the projection problem.
        MindtPy = self.working_model.MindtPy_utils
        config = self.config
        self.process_objective(update_var_con_list=config.add_regularization is None)
        # The epigraph constraint is very "flat" for branching rules.
        # If ROA/RLP-NLP is activated and the original objective function is linear, we will use the original objective for the main mip.
        if (
            MindtPy.objective_list[0].expr.polynomial_degree()
            in self.mip_objective_polynomial_degree
            and config.add_regularization is not None
        ):
            MindtPy.objective_list[0].activate()
            MindtPy.objective_constr.deactivate()
            MindtPy.objective.deactivate()
